"""Calibrate accelerometer data."""

import abc
import math
from collections.abc import Generator
from dataclasses import dataclass
from datetime import datetime
from typing import Literal, Union, cast

import numpy as np
import polars as pl
from scipy import optimize
from sklearn import linear_model
from sklearn import metrics as sklearn_metrics

from wristpy.core import computations, config, exceptions, models

logger = config.get_logger()


@dataclass
class LinearTransformation:
    """Data class that contains scale and offset values for calibration.

    Attributes:
        scale: 3-element 1-D array of floats used to scale accelerameter data for
            calibration. Each element corresponds to the scaling factor for each axis.
        offset: 3-element 1-D array of floats used to offset accelerameter data for
            calibration. Each element corresponds to the offset value for each axis.
    """

    scale: np.ndarray
    offset: np.ndarray


class AbstractCalibrator(abc.ABC):
    """Abstract class defining the interface for the different calibration methods."""

    @abc.abstractmethod
    def __init__(self) -> None:
        """Initialization function for the calibrator."""
        pass

    @abc.abstractmethod
    def run_calibration(self, acceleration: models.Measurement) -> models.Measurement:
        """The calibration method must contain a run_calibration function.

        The function must take the acceleration measurement object as input and return
        a measurement object that contains the calibrated acceleration data.
        """
        pass


class GgirCalibration(AbstractCalibrator):
    """Implements the GGIR calibration on accelerometer data.

    This class implements methods for autocalibrating accelerometer data using either
    entire dataset or subsets, as determined by the settings. Depending on the settings
    a scale and offset value is determined and applied to the data which, if
    successful, is returned to the user.


    Attributes:
        chunked: If true will perform calibration on subsets of data if false will
            calibrate on entire dataset.
        min_acceleration: Minimum acceleration for sphere criteria, in g-force.
        min_calibration_hours: Minimum hours of data required for calibration.
        no_motion_threshold: Minimum standard deviation for no-motion detection.
        max_iterations: Maximum number of iterations for optimization.
        error_tolerance: Tolerance for optimization convergence.
        max_calibration_error: Threshold for the maximum allowable calibration error.

    """

    def __init__(
        self,
        chunked: bool = False,
        min_acceleration: float = 0.3,
        min_calibration_hours: int = 72,
        no_motion_threshold: float = 0.013,
        max_iterations: int = 1000,
        error_tolerance: float = 1e-10,
        max_calibration_error: float = 0.01,
    ) -> None:
        """Initializes class.

        Attributes:
            chunked: Determines if either entire dataset will be used, or if
                calibration will be attempted with a subset of the data. Set to false by
                default, when true will initiate _chunked_calibration from the run()
                method instead of standard _calibration.
            min_acceleration: The value on either side of 0g for each axis.
                Determines if sphere is sufficiently populated to obtain meaningful
                calibration result. Default is 0.3g.
            min_calibration_hours: The minimum amount of data in hours needed to
                perform calibration. Default is 72. If chunked calibration is selected
                this will be the size of the initial subset taken for calibration. If
                error has not been reduced below the max_calibration_error value, an
                additional 12 hours will be taken until all data is used or calibration
                is successful.
            no_motion_threshold: The standard deviation criteria used to select
                portions of the data with no movement. Rolling windows with standard
                deviations below this value will be determined as "still". This value is
                likely to be different between devices. Default is 0.013g. This value
                was determined just above the empirically derived  baseline standard
                deviation due to noise (0.010g)(van Hees et al. 2014). If measuring the
                noise in a bench-top test, this threshold should be about `1.2 * noise`.
            max_iterations: The maximum amount of iterations for the closest_point_fit
                method. Default is 1000, generally should be left at this value.
            error_tolerance: Tolerated level of error, when the
                closest_point_fit method arrives at this value or better, the process
                ends. Default is 1e-10. Generally should be left at this value.
            max_calibration_error: Maximum acceptable calibration error. If calibration
                cannot reach this threshold it will error.

        Returns:
            None
        """
        self.chunked = chunked
        self.min_acceleration = min_acceleration
        self.min_calibration_hours = min_calibration_hours
        self.no_motion_threshold = no_motion_threshold
        self.max_iterations = max_iterations
        self.error_tolerance = error_tolerance
        self.max_calibration_error = max_calibration_error

    def run_calibration(self, acceleration: models.Measurement) -> models.Measurement:
        """Runs calibration on acceleration data based on settings.

        If the chunked argument is set to true, it will run calibration on an initial
        chunk of the data the size of which is set to min_calibration_hours. If it fails
        to calibrate based on this subset, it will add 12-hour chunks to the subset
        until calibration succeeds, or fails on the entire dataset. Otherwise, if
        chunked is false (default) calibration will be conducted on the whole of the
        dataset. Calibration is successful when scale and offset values, that
        sufficiently minimize error when applied to the data, are found. The scale and
        offset values are then applied to every data point in the dataset which is
        then returned as calibrated models.Measurement object.

        Args:
            acceleration: the accelerometer data containing x,y,z axis
                data and time stamps.

        Returns:
            A Measurement object that contains the calibrated acceleration data.

        Raises:
            ValueError: If the acceleration data does not meet the specified minimum
                hours of data as given by min_calibration_hours.
            SphereCriteriaError: If the sphere is not sufficiently populated, i.e. every
                axis does not have at least 1 value both above and below  the + and
                - value of min_acceleraiton.
            CalibrationError: If the calibration process fails to get below the
                `max_calibration_error` threshold.
        """
        logger.debug("Starting calibration.")
        data_range = cast(datetime, acceleration.time.max()) - cast(
            datetime, acceleration.time.min()
        )
        total_hours = math.floor(data_range.total_seconds() / 3600)

        if total_hours < self.min_calibration_hours:
            raise exceptions.CalibrationError(
                (
                    f"Calibration requires {self.min_calibration_hours} hours"
                    f"but only {total_hours} hours of data were given."
                ),
            )

        if self.chunked:
            linear_transformation = self._chunked_calibration(acceleration=acceleration)
        else:
            linear_transformation = self._calibrate(acceleration=acceleration)

        calibrated_acceleration = (
            acceleration.measurements * linear_transformation.scale
        ) + linear_transformation.offset

        return models.Measurement(
            measurements=calibrated_acceleration, time=acceleration.time
        )

    def _chunked_calibration(
        self, acceleration: models.Measurement
    ) -> LinearTransformation:
        """Chunks the data into subsets, to calibrate on smaller sections of data.

        The first chunk is determined by the min_calibration_hours. Should calibration
        fail on that subset of data, additional chunks of data are added to attempt
        calibration on. This continues until all data is tried, or calibration succeeds.
        Chunks are added in increments of 12 hours.

        Args:
            acceleration: the accelerometer data containing x,y,z axis
                data and time stamps.

        Returns:
            A LinearTransformation object with scale and offset attributes to be applied
            to acceleration data for calibration.

        Raises:
            CalibrationError: If all possible chunks have been used and the calibration
                process fails to get below the `max_calibration_error` threshold.
        """
        logger.debug("Running chunked calibration.")
        for chunk in self._get_chunk(acceleration):
            try:
                return self._calibrate(chunk)
            except (
                exceptions.SphereCriteriaError,
                exceptions.CalibrationError,
                exceptions.NoMotionError,
            ):
                pass
        raise exceptions.CalibrationError(
            "After all chunks of data used calibration has failed."
        )

    def _get_chunk(
        self, acceleration: models.Measurement
    ) -> Generator[models.Measurement, None, None]:
        """Takes a subset of acceleration data to be used for calibration.

        Args:
            acceleration: the accelerometer data containing x,y,z axis
                data and time stamps.

        Returns:
            The minimum hours of accelerometer data + additional 12 hour chunks of data
            every time the generator function is called.

        """
        logger.debug("Getting chunk.")
        sampling_rate = GgirCalibration._get_sampling_rate(timestamps=acceleration.time)
        min_samples = int(self.min_calibration_hours * 3600 * sampling_rate)
        chunk_size = int(12 * 3600 * sampling_rate)
        total_samples = len(acceleration.measurements)
        if min_samples == total_samples:
            sample_indices = [total_samples]
        else:
            sample_indices = list(range(min_samples, total_samples, chunk_size))
            if sample_indices[-1] != total_samples:
                sample_indices += [total_samples]

        for idx in sample_indices:
            yield models.Measurement(
                measurements=acceleration.measurements[:idx, :],
                time=acceleration.time[:idx],
            )

    def _calibrate(self, acceleration: models.Measurement) -> LinearTransformation:
        """Calibrates data and returns scale and offset values.

        The acceleration data is processed by the _extract_no_motion function, which
        returns the portions of the data where the subject was still. This data ideally
        should all have points with a norm of 1. If we take a unit sphere, all the
        points in no_motion_data should be found along it's surface. As this is not the
        case we calibrate by finding offset and scale values that will most closely make
        the whole of our data lie along this unit sphere. This is done by the
        _closest_point_fit function. We then transform the data by applying the scale
        and offset values we have found, and if the error has been sufficiently reduced,
        these values are returned. If not calibration fails and raises a calibration
        error.

        Args:
            acceleration: the accelerometer data containing x,y,z axis
                data and time stamps.

        Returns:
            A LinearTransformation object with scale and offset attributes to be applied
            to acceleration data for calibration.

        Raises:
            CalibrationError: If the calibration process fails to converge or the final
                error exceeds the `max_calibration_error` threshold.

        References:
            van Hees VT, Fang Z, Langford J, et al. Autocalibration of accelerometer
            data for free-living physical activity assessment using local gravity
            and temperature: an evaluation on four continents. J Appl Physiol (1985)
            2014 Oct 1;117(7):738-44. doi: 10.1152/japplphysiol.00421.2014.
        """
        logger.debug("Attempting to calibrate...")
        no_motion_data = _extract_no_motion(acceleration=acceleration)

        linear_transformation = self._closest_point_fit(no_motion_data=no_motion_data)

        no_motion_calibrated = (
            no_motion_data * linear_transformation.scale
        ) + linear_transformation.offset

        cal_error_initial = np.round(
            np.mean(abs(np.linalg.norm(no_motion_data, axis=1) - 1)), decimals=5
        )
        cal_error_end = np.around(
            np.mean(abs(np.linalg.norm(no_motion_calibrated, axis=1) - 1)), decimals=5
        )

        if (cal_error_end > cal_error_initial) or (
            cal_error_end >= self.max_calibration_error
        ):
            raise exceptions.CalibrationError(
                "Calibration error could not be sufficiently minimized. "
                f"Initial Error: {cal_error_initial}, Final Error: {cal_error_end}, "
                f"Error threshold: {self.max_calibration_error}"
            )
        logger.debug(
            "Calibration successful. Scale: %s, Offset: %s",
            linear_transformation.scale,
            linear_transformation.offset,
        )
        return linear_transformation

    def _closest_point_fit(self, no_motion_data: np.ndarray) -> LinearTransformation:
        """Applies closest point fit to no motion data to calibrated accelerometer data.

        This method implements an iterative algorithm that finds the optimal scale and
        offset for calibrating accelerometer data. The ndarray it processes contains the
        periods of "no motion" identified by the _extract_no_motion method. This data is
        then fit to an idealized unit sphere which represents the accelerometer data
        under noiseless conditions. The scale and offset derived from this process are
        used to calibrate our data along each axis.

        Args:
            no_motion_data: The acceleration data during periods of no
                motion, in order to determine scale and offset.

        Returns:
            A LinearTransformation object with scale and offset attributes to be applied
            to acceleration data for calibration.

        Raises:
            SphereCriteriaError: If the sphere is not sufficiently populated, i.e. every
                axis does not have at least 1 value both above and below  the + and
                - value of min_acceleraiton.
            ZeroScaleError: Numerical instability inherent to the algorithm may cause
                values of the scale vector to tend toward zero under certain conditions.
                This will cause calibration to fail.

        References:
            van Hees VT, Fang Z, Langford J, et al. Autocalibration of accelerometer
            data for free-living physical activity assessment using local gravity
            and temperature: an evaluation on four continents. J Appl Physiol (1985)
            2014 Oct 1;117(7):738-44. doi: 10.1152/japplphysiol.00421.2014.
        """
        logger.debug("Beginning closest point fit.")
        sphere_criteria_check = np.all(
            (no_motion_data.min(axis=0) < -self.min_acceleration)
            & (no_motion_data.max(axis=0) > self.min_acceleration)
        )
        if not sphere_criteria_check:
            raise exceptions.SphereCriteriaError(
                "Did not meet criteria to sufficiently populate sphere"
            )

        weights = np.ones(no_motion_data.shape[0]) * 100
        previous_residual = np.inf

        linear_regression_model = linear_model.LinearRegression()

        offset = np.zeros(3)
        scale = np.ones(3)
        for _ in range(self.max_iterations):
            current = (no_motion_data * scale) + offset
            closest_point = current / np.linalg.norm(current, axis=1, keepdims=True)
            linear_regression_model.fit(current, closest_point, sample_weight=weights)
            offset_change = linear_regression_model.intercept_
            scale_change = np.diag(linear_regression_model.coef_)

            if np.all((scale * scale_change) < 1e-8):
                raise exceptions.CalibrationError(
                    """Calibration has failed as a result of zero scale values."""
                )

            scale *= scale_change
            offset += offset_change / scale

            residual = 0.03 * sklearn_metrics.mean_squared_error(
                current, closest_point, sample_weight=weights
            )

            weights = np.minimum(
                1 / np.linalg.norm(current - closest_point, axis=1), 100
            )

            logger.debug("Scale: %s, Offset: %s, Residual: %s", scale, offset, residual)
            if abs(residual - previous_residual) < self.error_tolerance:
                logger.debug("Change in residual below error tolerance, ending loop.}")
                break

            previous_residual = residual

        return LinearTransformation(scale=scale, offset=offset)

    @staticmethod
    def _get_sampling_rate(timestamps: pl.Series) -> int:
        """Get the sampling rate.

        Args:
            timestamps: polars series of datetime objects representing the time points
            of each sample in the acceleration data.

        Returns:
            sampling rate in Hz.
        """
        sampling_rate = timestamps.len() / round(
            (
                cast(datetime, timestamps.max()) - cast(datetime, timestamps.min())
            ).total_seconds()
        )

        return round(sampling_rate)


class ConstrainedMinimizationCalibration(AbstractCalibrator):
    """Calibrates accelerometer data using the default wristpy method.

    This is a modification of the method proposed by Van Hees et al. (2014), in which we
    use all available data, make use of the scipy optimize minimize function to solve
    for the scaling and offset parameters of a LinearTransformation that will be applied
    to the data. The minimization function looks to minimize the error between the
    no_motion_data and the unit sphere.

    Attributes:
        acceleration: The accelerometer data that we want to calibrate.
        no_motion_threshold: Minimum standard deviation for no-motion detection.
    """

    def __init__(
        self,
        no_motion_threshold: float = 0.013,
        max_iterations: int = 1000,
        max_calibration_error: float = 0.01,
    ) -> None:
        """Initializes class.

        Args:
            no_motion_threshold: The standard deviation criteria used to find
                periods of no motion. Default is 0.013g.
            max_iterations: The maximum amount of iterations for the
                closest_point_fit function.
            max_calibration_error: Maximum acceptable calibration error. If calibration
                cannot reach this threshold it will error.

        Returns:
            None
        """
        self.no_motion_threshold = no_motion_threshold
        self.max_iterations = max_iterations
        self.max_calibration_error = max_calibration_error

    def run_calibration(self, acceleration: models.Measurement) -> models.Measurement:
        """Runs calibration on acceleration data.

        Args:
            acceleration: the accelerometer data containing x,y,z axis
                data and time stamps.

        Returns:
            A Measurement object that contains the calibrated acceleration data.

        Raises:
            CalibrationError: If the calibration process fails to converge or the final
                error exceeds the `max_calibration_error` threshold.
        """
        logger.debug("Starting calibration.")
        no_motion_data = _extract_no_motion(
            acceleration=acceleration,
            no_motion_threshold=self.no_motion_threshold,
        )
        linear_transformation = self._closest_point_fit(no_motion_data=no_motion_data)

        logger.debug(
            "Calibration successful. Scale: %s, Offset: %s. ",
            linear_transformation.scale,
            linear_transformation.offset,
        )

        calibrated_acceleration = (
            acceleration.measurements * linear_transformation.scale
        ) + linear_transformation.offset

        return models.Measurement(
            measurements=calibrated_acceleration, time=acceleration.time
        )

    def _closest_point_fit(self, no_motion_data: np.ndarray) -> LinearTransformation:
        """Find linear transformation parameters that minimizes distance to unit sphere.

        This function implements the scipy optimize.minimize function to find the
        optimal scale and offset parameters that will minimize the error function,
        which is defined as the distance of the no_motion acceleration data
        from the unit sphere (where the unit sphere represents the ideal data points
        under 1g acceleration due to Earth's gravity). The initial guess for the
        scale and offset are chosen as 1/0, we provide constrained bounds for the
        scale and offset parameters to avoid the case where the scale can be
        set to 0 and offset to 1/sqrt(3) (exact unit sphere).

        Args:
            no_motion_data: The acceleration data during periods of no motion.

        Returns:
            A LinearTransformation object with scale and offset parameters to be applied
            to acceleration data for calibration.

        Raises:
            CalibrationError: If the optimization process fails to converge or if the
                calibration error is not sufficiently minimized.
        """
        start_scale = np.ones(3)
        start_offset = np.zeros(3)

        def get_distance_to_unit_sphere(params: list[float]) -> float:
            scale = params[:3]
            offset = params[3:]
            distances = np.linalg.norm((no_motion_data * scale) + offset, axis=1) - 1
            return np.sum(distances**2)

        initial_guess = np.concatenate([start_scale, start_offset])
        data_range = no_motion_data.max() - no_motion_data.min()
        scale_bounds = [(0.1, None)] * 3
        offset_bounds = [(-0.5 * data_range, 0.5 * data_range)] * 3
        bounds = scale_bounds + offset_bounds
        result = optimize.minimize(
            get_distance_to_unit_sphere,
            initial_guess,
            options={"maxiter": self.max_iterations},
            bounds=bounds,
        )

        if result.success:
            optimal_scale = result.x[:3]
            optimal_offset = result.x[3:]
            cal_error_end = np.sqrt(result.fun / len(no_motion_data))
        else:
            raise exceptions.CalibrationError("Optimization failed.")

        if cal_error_end >= self.max_calibration_error:
            cal_error_initial = np.mean(
                (np.linalg.norm(no_motion_data, axis=1) - 1) ** 2
            )
            logger.debug(
                "Calibration error could not be sufficiently minimized."
                "Initial Error: %s,  Final Error: %s, Error threshold: %s.",
                cal_error_initial,
                cal_error_end,
                self.max_calibration_error,
            )
            raise exceptions.CalibrationError(
                "Calibration error could not be sufficiently minimized."
            )

        return LinearTransformation(scale=optimal_scale, offset=optimal_offset)


def _extract_no_motion(
    acceleration: models.Measurement, no_motion_threshold: float = 0.013
) -> np.ndarray:
    """Identifies areas of stillness using standard deviation and mean.

    The function first takes a moving standard deviation and a moving mean of the
    acceleration data in 10 second epochs. These ndarrays are then used to identify
    the portions of the data that have a standard deviation below
    no_motion_threshold and a mean value < 2. These epochs are determined to be
    the periods where the accelerometer was influenced by no motion beyond the
    force of gravity. If periods of no motion are found, the ndarray is returned,
    to be fit to the idealized unit sphere for the purposes of calibration.

    Args:
        acceleration: the accelerometer data containing x,y,z axis
            data and time stamps.
        no_motion_threshold: Threshold for the standard deviaton of acceleration data
            used to find periods of no motion.

    Returns:
        An ndarray containing the accelerometer data determined to have no motion.

    Raises:
        NoMotionError: If no portions of data meet no motion criteria as defined by
            no_motion_check.

    References:
        van Hees VT, Fang Z, Langford J, et al. Autocalibration of accelerometer
        data for free-living physical activity assessment using local gravity
        and temperature: an evaluation on four continents. J Appl Physiol (1985)
        2014 Oct 1;117(7):738-44. doi: 10.1152/japplphysiol.00421.2014.
    """
    logger.debug("Extracting no motion.")
    moving_sd = computations.moving_std(acceleration, 10)
    moving_mean = computations.moving_mean(acceleration, 10)
    no_motion_check = np.all(
        moving_sd.measurements < no_motion_threshold, axis=1
    ) & np.all(np.abs(moving_mean.measurements) < 2, axis=1)

    no_motion_data = moving_mean.measurements[no_motion_check]

    if not np.any(no_motion_data):
        raise exceptions.NoMotionError(
            "Zero non-motion epochs found. Data did not meet criteria."
        )

    return no_motion_data


class CalibrationDispatcher:
    """Class used to select and implement appropriate calibrator."""

    _calibrator: Union[GgirCalibration, ConstrainedMinimizationCalibration]

    def __init__(self, name: Literal["ggir", "gradient"]) -> None:
        """Initializes the calibrator to one of the predefined calibrators.

        Args:
            name: The name of the calibrator to use. Options are "ggir" or "gradient".

        Raises:
            ValueError: If an unknown calibrator is given
        """
        if name == "ggir":
            self._calibrator = GgirCalibration()
        elif name == "gradient":
            self._calibrator = ConstrainedMinimizationCalibration()
        else:
            raise ValueError("Unknown calibrator.")

    def run(
        self, acceleration: models.Measurement, *, return_input_on_error: bool = False
    ) -> models.Measurement:
        """Runs calibration on acceleration data.

        Args:
            acceleration: the accelerometer data containing x,y,z axis
                data and time stamps.
            return_input_on_error: If true, will return the input acceleration data
                if calibration fails. If false, will raise an exception.

        Returns:
            A Measurement object that contains the calibrated acceleration data.

        Raises:
            CalibrationError: If the calibration process fails to converge or the final
                error exceeds the `max_calibration_error` threshold.
            SphereCriteriaError: If the sphere is not sufficiently populated, i.e. every
                axis does not have at least 1 value both above and below  the + and
                - value of min_acceleraiton.
            NoMotionError: If no portions of data meet no motion criteria as defined by
                no_motion_check.
        """
        try:
            return self._calibrator.run_calibration(acceleration)
        except (
            exceptions.CalibrationError,
            exceptions.SphereCriteriaError,
            exceptions.NoMotionError,
        ) as exc_info:
            if not return_input_on_error:
                raise
            logger.warning(
                "Calibration FAILED: %s. Proceeding without calibration.", exc_info
            )
            return acceleration



"""Python based runner."""

import itertools
import logging
import pathlib
from typing import Dict, Literal, Optional, Sequence, Tuple, Union

from wristpy.core import config, exceptions, models
from wristpy.io.readers import readers
from wristpy.io.writers import writers
from wristpy.processing import (
    analytics,
    calibration,
    idle_sleep_mode_imputation,
    metrics,
    nonwear_utils,
)

logger = config.get_logger()

VALID_FILE_TYPES = (".csv", ".parquet")


def run(
    input: Union[pathlib.Path, str],
    output: Optional[Union[pathlib.Path, str]] = None,
    thresholds: Optional[Tuple[float, float, float]] = None,
    calibrator: Union[
        None,
        Literal["ggir", "gradient"],
    ] = "gradient",
    epoch_length: float = 5,
    activity_metric: Literal["enmo", "mad", "ag_count", "mims"] = "enmo",
    nonwear_algorithm: Sequence[Literal["ggir", "cta", "detach"]] = ["ggir"],
    verbosity: int = logging.WARNING,
    output_filetype: Optional[Literal[".csv", ".parquet"]] = None,
) -> Union[writers.OrchestratorResults, Dict[str, writers.OrchestratorResults]]:
    """Runs main processing steps for wristpy on single files, or directories.

    The run() function will execute the run_file() function on individual files, or
    run_directory() on entire directories. When the input path points to a file, the
    name of the save file will be taken from the given output path (if any). When the
    input path points to a directory the output path must be a valid directory as well.
    Output file names will be derived from original file names in the case of directory
    processing.


    Args:
        input: Path to the input file or directory of files to be read. Currently,
            this supports .bin and .gt3x
        output: Path to directory data will be saved to. If processing a single file the
            path should end in the save file name in either .csv or .parquet formats.
        thresholds: The cut points for the light, moderate, and vigorous thresholds,
            given in that order. Values must be asscending, unique, and greater than 0.
            Default values are optimized for subjects ages 7-11 [1][3].
        calibrator: The calibrator to be used on the input data.
        epoch_length: The temporal resolution in seconds, the data will be down sampled
            to. It must be > 0.0.
        activity_metric: The metric to be used for physical activity categorization.
        nonwear_algorithm: The algorithm to be used for nonwear detection.
        verbosity: The logging level for the logger.
        output_filetype: Specifies the data format for the save files. Must be None when
            processing files, must be a valid file type when processing directories.

    Returns:
        All calculated data in a save ready format as a Results object or as a
        dictionary of OrchestratorResults objects.

    Raises:
        ValueError: If the physical activity thresholds are not unique or not in
            ascending order.
        ValueError: If processing a file and the output_filetype is not None
        ValueError: If output is None but output_filetype is not None.

    References:
        [1] Hildebrand, M., et al. (2014). Age group comparability of raw accelerometer
        output from wrist- and hip-worn monitors. Medicine and Science in Sports and
        Exercise, 46(9), 1816-1824.
        [2] Treuth MS, Schmitz K, Catellier DJ, McMurray RG, Murray DM, Almeida MJ,
        Going S, Norman JE, Pate R. Defining accelerometer thresholds for activity
        intensities in adolescent girls. Med Sci Sports Exerc. 2004 Jul;36(7):1259-66.
        PMID: 15235335; PMCID: PMC2423321.
        [3] Karas M, Muschelli J, Leroux A, Urbanek J, Wanigatunga A, Bai J,
        Crainiceanu C, Schrack J Comparison of Accelerometry-Based Measures of Physical
        Activity: Retrospective Observational Data Analysis Study JMIR Mhealth Uhealth
        2022;10(7):e38077 URL: https://mhealth.jmir.org/2022/7/e38077 DOI: 10.2196/38077
    """
    logger.setLevel(verbosity)

    input = pathlib.Path(input)
    output = pathlib.Path(output) if output is not None else None

    if activity_metric == "enmo":
        thresholds = thresholds or (0.0563, 0.1916, 0.6958)
    elif activity_metric == "mad":
        thresholds = thresholds or (0.029, 0.338, 0.604)
    elif activity_metric == "ag_count":
        thresholds = thresholds or (100, 3000, 5200)
    elif activity_metric == "mims":
        thresholds = thresholds or (10.558, 15.047, 19.614)

    if not (0 <= thresholds[0] < thresholds[1] < thresholds[2]):  # type: ignore
        message = "Threshold values must be >=0, unique, and in ascending order."
        logger.error(message)
        raise ValueError(message)

    if input.is_file():
        if output_filetype is not None:
            raise ValueError(
                "When processing single files, output_filetype should be None - "
                "the file type will be determined from the output path."
            )
        logger.debug("Input is file, forwarding to run_file with output=%s", output)

        return _run_file(
            input=input,
            output=output,
            thresholds=thresholds,
            calibrator=calibrator,
            epoch_length=epoch_length,
            activity_metric=activity_metric,
            verbosity=verbosity,
            nonwear_algorithm=nonwear_algorithm,
        )

    return _run_directory(
        input=input,
        output=output,
        thresholds=thresholds,
        calibrator=calibrator,
        epoch_length=epoch_length,
        activity_metric=activity_metric,
        verbosity=verbosity,
        output_filetype=output_filetype,
        nonwear_algorithm=nonwear_algorithm,
    )


def _run_directory(
    input: pathlib.Path,
    output: Optional[pathlib.Path] = None,
    thresholds: Tuple[float, float, float] = (0.0563, 0.1916, 0.6958),
    calibrator: Union[
        None,
        Literal["ggir", "gradient"],
    ] = "gradient",
    epoch_length: float = 5,
    nonwear_algorithm: Sequence[Literal["ggir", "cta", "detach"]] = ["ggir"],
    verbosity: int = logging.WARNING,
    output_filetype: Optional[Literal[".csv", ".parquet"]] = None,
    activity_metric: Literal["enmo", "mad", "ag_count", "mims"] = "enmo",
) -> Dict[str, writers.OrchestratorResults]:
    """Runs main processing steps for wristpy on  directories.

    The run_directory() function will execute the run_file() function on entire
    directories. The input and output (if any) paths must directories. An
    output_filetype must be specified if and only if an output is given. Output file
    names will be derived from input file names.


    Args:
        input: Path to the input directory of files to be read. Currently,
            this supports .bin and .gt3x
        output: Path to directory data will be saved to.
        thresholds: The cut points for the light, moderate, and vigorous thresholds,
            given in that order. Values must be asscending, unique, and greater than 0.
            Default values are optimized for subjects ages 7-11 [1][2].
        calibrator: The calibrator to be used on the input data.
        epoch_length: The temporal resolution in seconds, the data will be down sampled
            to. It must be > 0.0.
        nonwear_algorithm: The algorithm to be used for nonwear detection.
        verbosity: The logging level for the logger.
        output_filetype: Specifies the data format for the save files.
        activity_metric: The metric to be used for physical activity categorization.

    Returns:
        All calculated data in a save ready format as a dictionary of
        OrchestratorResults objects.

    Raises:
        ValueError: The output given is not a directory.
        ValueError: The output_filetype is not a valid type.
        FileNotFoundError: If the input directory contained no files of a valid type.


    References:
        [1] Hildebrand, M., et al. (2014). Age group comparability of raw accelerometer
        output from wrist- and hip-worn monitors. Medicine and Science in Sports and
        Exercise, 46(9), 1816-1824.
        [2] Karas M, Muschelli J, Leroux A, Urbanek J, Wanigatunga A, Bai J,
        Crainiceanu C, Schrack J Comparison of Accelerometry-Based Measures of Physical
        Activity: Retrospective Observational Data Analysis Study JMIR Mhealth Uhealth
        2022;10(7):e38077 URL: https://mhealth.jmir.org/2022/7/e38077 DOI: 10.2196/38077
    """
    if output is None and output_filetype is not None:
        raise ValueError("If no output is given, output_filetype must be None.")

    if output is not None:
        if output.is_file():
            raise ValueError(
                "Output is a file, but must be a directory when input is a directory."
            )
        if output_filetype not in VALID_FILE_TYPES:
            raise ValueError(
                "Invalid output_filetype: "
                f"{output_filetype}. Valid options are: {VALID_FILE_TYPES}."
            )

    file_names = list(itertools.chain(input.glob("*.gt3x"), input.glob("*.bin")))

    if not file_names:
        raise FileNotFoundError(f"Directory {input} contains no .gt3x or .bin files.")

    results_dict = {}
    for file in file_names:
        output_file_path = (
            output / pathlib.Path(file.stem).with_suffix(output_filetype)  # type: ignore[arg-type] # if output is defined, so is output_filetype
            if output
            else None
        )
        logger.debug(
            "Processing directory: %s, current file: %s, save path: %s",
            input,
            file,
            output_file_path,
        )
        try:
            results_dict[str(file)] = _run_file(
                input=input / file,
                output=output_file_path,
                thresholds=thresholds,
                calibrator=calibrator,
                epoch_length=epoch_length,
                verbosity=verbosity,
                nonwear_algorithm=nonwear_algorithm,
                activity_metric=activity_metric,
            )
        except Exception as e:
            logger.error("Did not run file: %s, Error: %s", file, e)
    logger.info("Processing for directory %s completed successfully.", output)
    return results_dict


def _run_file(
    input: pathlib.Path,
    output: Optional[pathlib.Path] = None,
    thresholds: Tuple[float, float, float] = (0.0563, 0.1916, 0.6958),
    calibrator: Union[
        None,
        Literal["ggir", "gradient"],
    ] = "gradient",
    epoch_length: float = 5.0,
    activity_metric: Literal["enmo", "mad", "ag_count", "mims"] = "enmo",
    nonwear_algorithm: Sequence[Literal["ggir", "cta", "detach"]] = ["ggir"],
    verbosity: int = logging.WARNING,
) -> writers.OrchestratorResults:
    """Runs main processing steps for wristpy and returns data for analysis.

    The run_file() function will provide the user with the specified physical activity
    metric, anglez, physical activity levels, detected sleep periods, and nonwear data.
    All measures will be in the same temporal resolution.
    Users may choose from 'ggir' and 'gradient' calibration methods,
    or enter None to proceed without calibration.

    Args:
        input: Path to the input file to be read. Currently, this supports .bin and
            .gt3x
        output: Path to save data to. The path should end in the save file name in
            either .csv or .parquet formats.
        thresholds: The cut points for the light, moderate, and vigorous thresholds,
            given in that order. Values must be ascending, unique, and greater than 0.
            Default values are optimized for subjects ages 7-11 [1] - [3].
        calibrator: The calibrator to be used on the input data.
        epoch_length: The temporal resolution in seconds, the data will be down sampled
            to. It must be > 0.0.
        activity_metric: The metric to be used for physical activity categorization.
        nonwear_algorithm: The algorithm to be used for nonwear detection. A sequence of
            algorithms can be provided. If so, a majority vote will be taken.
        verbosity: The logging level for the logger.

    Returns:
        All calculated data in a save ready format as a OrchestratorResults object.

    Raises:
        ValueError: If an invalid Calibrator is chosen.
        ValueError: If epoch_length is not greater than 0.

    References:
        [1] Hildebrand, M., et al. (2014). Age group comparability of raw accelerometer
        output from wrist- and hip-worn monitors. Medicine and Science in Sports and
        Exercise, 46(9), 1816-1824.
        [2] Aittasalo, M., Vähä-Ypyä, H., Vasankari, T. et al. Mean amplitude deviation
        calculated from raw acceleration data: a novel method for classifying the
        intensity of adolescents' physical activity irrespective of accelerometer brand.
        BMC Sports Sci Med Rehabil 7, 18 (2015). https://doi.org/10.1186/s13102-015-0010-0.
        [3] Karas M, Muschelli J, Leroux A, Urbanek J, Wanigatunga A, Bai J,
        Crainiceanu C, Schrack J Comparison of Accelerometry-Based Measures of Physical
        Activity: Retrospective Observational Data Analysis Study JMIR Mhealth Uhealth
        2022;10(7):e38077 URL: https://mhealth.jmir.org/2022/7/e38077 DOI: 10.2196/38077
    """
    logger.setLevel(verbosity)
    if output is not None:
        writers.OrchestratorResults.validate_output(output=output)

    parameters_dictionary = {
        "thresholds": list(thresholds),
        "calibrator": calibrator,
        "epoch_length": epoch_length,
        "activity_metric": activity_metric,
        "nonwear_algorithm": list(nonwear_algorithm),
        "input_file": str(input),
    }

    if calibrator is not None and calibrator not in ["ggir", "gradient"]:
        msg = (
            f"Invalid calibrator: {calibrator}. Choose: 'ggir', 'gradient'. "
            "Enter None if no calibration is desired."
        )
        logger.error(msg)
        raise ValueError(msg)

    if epoch_length <= 0:
        msg = "Epoch_length must be greater than 0."
        logger.error(msg)
        raise ValueError(msg)

    watch_data = readers.read_watch_data(input)

    if calibrator is None:
        logger.debug("Running without calibration")
        calibrated_acceleration = watch_data.acceleration
    else:
        calibrated_acceleration = calibration.CalibrationDispatcher(calibrator).run(
            watch_data.acceleration, return_input_on_error=True
        )

    if watch_data.idle_sleep_mode_flag:
        logger.debug("Imputing idle sleep mode gaps.")
        calibrated_acceleration = (
            idle_sleep_mode_imputation.impute_idle_sleep_mode_gaps(
                calibrated_acceleration
            )
        )

    anglez = metrics.angle_relative_to_horizontal(
        calibrated_acceleration, epoch_length=epoch_length
    )
    activity_measurement = _compute_activity(
        calibrated_acceleration,
        activity_metric,
        epoch_length,
        dynamic_range=watch_data.dynamic_range,
    )

    sleep_detector = analytics.GgirSleepDetection(anglez)
    sleep_windows = sleep_detector.run_sleep_detection()

    nonwear_array = nonwear_utils.get_nonwear_measurements(
        calibrated_acceleration=calibrated_acceleration,
        temperature=watch_data.temperature,
        non_wear_algorithms=nonwear_algorithm,
    )

    nonwear_epoch = nonwear_utils.nonwear_array_cleanup(
        nonwear_array=nonwear_array,
        reference_measurement=activity_measurement,
        epoch_length=epoch_length,
    )

    physical_activity_levels = analytics.compute_physical_activty_categories(
        activity_measurement, thresholds
    )

    sleep_array = analytics.sleep_cleanup(
        sleep_windows=sleep_windows, nonwear_measurement=nonwear_epoch
    )
    results = writers.OrchestratorResults(
        physical_activity_metric=activity_measurement,
        anglez=anglez,
        physical_activity_levels=physical_activity_levels,
        sleep_status=sleep_array,
        nonwear_status=nonwear_epoch,
        processing_params=parameters_dictionary,
    )
    if output is not None:
        try:
            results.save_results(output=output)
        except (
            exceptions.InvalidFileTypeError,
            PermissionError,
            FileExistsError,
        ) as exc_info:
            # Allowed to pass to recover in Jupyter Notebook scenarios.
            logger.error(
                (
                    "Could not save output due to: %s. Call save_results "
                    "on the output object with a correct filename to save these "
                    "results.",
                    exc_info,
                )
            )
    logger.info("Processing for %s completed successfully.", input.stem)
    return results


def _compute_activity(
    acceleration: models.Measurement,
    activity_metric: Literal["ag_count", "mad", "enmo", "mims"],
    epoch_length: float,
    dynamic_range: Optional[tuple[float, float]],
) -> models.Measurement:
    """This is a helper function to organize the computation of the activity metric.

    This function organizes the logic for computing the requested physical activity
    metric at the desired temporal resolution.

    Args:
        acceleration: The acceleration data to compute the activity metric from.
        activity_metric: The metric to be used for physical activity categorization.
        epoch_length: The temporal resolution in seconds, the data will be down sampled
            to.
        dynamic_range: Tuple of the minimum and maximum accelerometer values. This
            argument is only relevant to the mims metric. Values are taken from watch
            metadata, if no metadata could be extracted, the default
            values of (-8,8) are used.

    Returns:
        A Measurement object with the computed physical activity metric.
    """
    if activity_metric == "ag_count":
        return metrics.actigraph_activity_counts(
            acceleration,
            epoch_length=epoch_length,
        )
    elif activity_metric == "mad":
        return metrics.mean_amplitude_deviation(acceleration, epoch_length=epoch_length)
    elif activity_metric == "mims":
        if dynamic_range is None:
            return metrics.monitor_independent_movement_summary_units(
                acceleration,
                epoch=epoch_length,
            )
        return metrics.monitor_independent_movement_summary_units(
            acceleration, epoch=epoch_length, dynamic_range=dynamic_range
        )

    return metrics.euclidean_norm_minus_one(acceleration, epoch_length=epoch_length)



"""Handle idle sleep mode special case."""

import numpy as np
import polars as pl

from wristpy.core import models


def impute_idle_sleep_mode_gaps(acceleration: models.Measurement) -> models.Measurement:
    """This function imputes the gaps in the idle sleep mode data.

    Gaps in the acceleration data are filled by assuming the watch is idle in a face up
    position. The acceleration data is filled in at a linear sampling rate, estimated
    based on the first 100 samples timestamps, with (0, 0, -1).

    In cases when the sampling rate leads to unevenly spaced samples within one second,
    e.g. 30Hz sampling rate has samples spaced at 33333333ns and 33333343ns within one
    second, the entire data set will be resampled at the highest effective sampling rate
    that allows for linearly spaced samples within one second,
    to nanosecond precision.

    Args:
        acceleration: The raw acceleration data.

    Returns:
        A Measurement object with the modified acceleration data.
    """

    def _find_effective_sampling_rate(sampling_rate: int) -> int:
        """Helper function to find the effective sampling rate.

        This function finds the new sampling rate that allows for linearly spaced
        samples within one second, to nanosecond precision.

        Args:
            sampling_rate: The original sampling rate.

        Returns:
            The new effective sampling rate.
        """
        for effective_sr in range(sampling_rate, 1, -1):
            if 1e9 % (1e9 / effective_sr) == 0:
                return effective_sr
        return 1

    acceleration_polars_df = pl.DataFrame(
        {
            "X": acceleration.measurements[:, 0],
            "Y": acceleration.measurements[:, 1],
            "Z": acceleration.measurements[:, 2],
            "time": acceleration.time,
        }
    )

    sampling_space_nanosec = np.mean(
        acceleration.time[:100]
        .diff()
        .drop_nulls()
        .dt.total_nanoseconds()
        .to_numpy()
        .astype(dtype=float)
    )

    sampling_rate = int(1e9 / sampling_space_nanosec)

    effective_sampling_rate = _find_effective_sampling_rate(sampling_rate)
    effective_sampling_interval = int(1e9 / effective_sampling_rate)

    filled_acceleration = (
        acceleration_polars_df.set_sorted("time")
        .group_by_dynamic("time", every=f"{effective_sampling_interval}ns")
        .agg(pl.exclude("time").mean())
        .upsample("time", every=f"{effective_sampling_interval}ns", maintain_order=True)
        .with_columns(
            pl.col("X").fill_null(value=0),
            pl.col("Y").fill_null(value=0),
            pl.col("Z").fill_null(value=-1),
        )
    )

    return models.Measurement.from_data_frame(filled_acceleration)



"""Calculate Monitor Independent Movement Summary Units."""

from typing import List, Literal, Optional, Tuple

import numpy as np
import polars as pl
from scipy import integrate, interpolate, signal, stats

from wristpy.core import config, models

logger = config.get_logger()


def interpolate_measure(
    acceleration: models.Measurement, new_frequency: int = 100
) -> models.Measurement:
    """Interpolate the measure to a new sampling rate using natural cubic spline.

    Args:
        acceleration: Accelerometer data and associated timestamps.
        new_frequency: The new frequency the measure will be interpolated to in Hz. For
            the purposes of the MIMS algorithm defaults to 100Hz.

    Returns:
        A Measurement object with interpolated acceleration data.
    """
    epoch_time_seconds = acceleration.time.dt.epoch(time_unit="ns").to_numpy() / 1e9
    start_time = epoch_time_seconds[0]
    end_time = epoch_time_seconds[-1]

    duration_s = end_time - start_time
    n_points = int(duration_s * new_frequency) + 1

    interpolated_time = np.linspace(start_time, end_time, n_points, endpoint=True)
    interpolated_data = np.zeros((len(interpolated_time), 3))

    for axis in range(3):
        cubic_spline = interpolate.CubicSpline(
            epoch_time_seconds, acceleration.measurements[:, axis], bc_type="natural"
        )
        interpolated_data[:, axis] = cubic_spline(interpolated_time)

    new_time_ns = (interpolated_time * 1e9).astype(np.int64)
    new_time_series = pl.Series(
        "interpolated_time", new_time_ns, dtype=pl.Datetime("ns")
    )

    return models.Measurement(measurements=interpolated_data, time=new_time_series)


def extrapolate_points(
    acceleration: models.Measurement,
    dynamic_range: Tuple[float, float] = (-8.0, 8.0),
    noise: float = 0.03,
    smoothing: float = 0.6,
    scale: float = 1,
    target_probability: float = 0.95,
    confidence_threshold: float = 0.5,
    neighborhood_size: float = 0.05,
    sampling_rate: int = 100,
) -> models.Measurement:
    """Identify maxed out values, and extrapolate points.

    Args:
        acceleration: Acceleration data that has been interpolated to 100Hz.
        dynamic_range: Dynamic range of device used. This information can be gathered
            from device metadata.
        noise: Device noise level.
        smoothing: Smoothing parameter for spline function. A value of 0 fits the data
            exactly, higher values will fit the data more smoothly.
        scale: Theta value, scale parameter for gamma distribution.
        target_probability: Threshold used in determining the shape parameter (k) of the
            gamma distribution. Probability that a value 3 standard deviations from the
            buffer zone is maxed-out.
        confidence_threshold: Threshold for what constitutes a major jump or drop in
            value.
        neighborhood_size: Duration of neighborhood in seconds. This parameter
            determines how much data on each side of the maxed-out region is used
            for fitting the local regression model. The parameter is used to calculate
            the number of points to create in the oversampled data (n_over).
        sampling_rate: The sampling rate in Hz. Assumed to take place after
            interpolation so it will typically be 100.

    Returns:
        A Measurement object with maxed out regions replaced with extrapolated values.

    References:
        John, D., Tang, Q., Albinali, F. and Intille, S., 2019. An Open-Source
            Monitor-Independent Movement Summary for Accelerometer Data Processing.
            Journal for the Measurement of Physical Behaviour, 2(4), pp.268-281.
    """
    time_numeric = acceleration.time.dt.epoch(time_unit="ns").to_numpy()
    extrapolated_axes = []
    for axis in acceleration.measurements.T:
        marker = _find_markers(
            axis=axis,
            dynamic_range=dynamic_range,
            noise=noise,
            scale=scale,
            target_probability=target_probability,
        )
        neighbors = _extrapolate_neighbors(
            marker=marker,
            confidence_threshold=confidence_threshold,
            neighborhood_size=neighborhood_size,
            sampling_rate=sampling_rate,
        )
        points = _extrapolate_fit(
            axis=axis,
            time_numeric=time_numeric,
            marker=marker,
            neighbors=neighbors,
            smoothing=smoothing,
            neighborhood_size=neighborhood_size,
            sampling_rate=sampling_rate,
        )
        extrapolated_values = _extrapolate_interpolate(
            axis=axis,
            time_numeric=time_numeric,
            marker=marker,
            points=points,
            confidence_threshold=confidence_threshold,
        )
        extrapolated_axes.append(extrapolated_values)

    extrapolated_acceleration = np.column_stack(extrapolated_axes)

    return models.Measurement(
        measurements=extrapolated_acceleration, time=acceleration.time
    )


def _find_markers(
    axis: np.ndarray,
    dynamic_range: Tuple[float, float],
    noise: float = 0.03,
    scale: float = 1.0,
    target_probability: float = 0.95,
) -> np.ndarray:
    """Determine the likelihood that each sample is near the limits of the device range.

    Args:
        axis: Acceleration data along one axis.
        dynamic_range: Tuple of floats indicating the device's value range. This
            information can be gathered from watch metadata.
        noise: Typical noise value for device.
        scale: Scale value (theta) for gamma distribution. Typically set at 1.
        target_probability: Threshold used in determining the shape parameter (k) of the
            gamma distribution. Probability that a value 3 standard deviations from the
            buffer zone is maxed-out.

    Returns:
        Array indicating the likelihood that corresponding value in the axis data is
        'maxed-out'.
    """
    positive_buffer_zone = max(dynamic_range) - 5 * (noise + 1e-5)
    negative_buffer_zone = min(dynamic_range) + 5 * (noise + 1e-5)
    shape_k = _brute_force_k(
        standard_deviation=3 * (noise + 1e-5),
        scale=scale,
        target_probability=target_probability,
    )
    marker = np.zeros(len(axis))
    positive_idx = axis >= 0
    negative_idx = axis < 0

    marker[positive_idx] = stats.gamma.cdf(
        axis[positive_idx] - positive_buffer_zone, a=shape_k, scale=scale
    )
    marker[negative_idx] = -stats.gamma.cdf(
        -axis[negative_idx] + negative_buffer_zone, a=shape_k, scale=scale
    )

    return marker


def _brute_force_k(
    standard_deviation: float,
    k_max: float = 0.5,
    k_min: float = 0.001,
    k_step: float = 0.001,
    scale: float = 1.0,
    target_probability: float = 0.95,
) -> float:
    """Find the shape parameter(k) for the gamma distribution.

    Args:
        standard_deviation: The point at which to evaluate the gamma cumulative
            distribution function, should be 3 times the noise.
        k_max: Maximum value for the shape parameter search range.
            Default is 0.5.
        k_min: Minimum value for the shape parameter search range.
            Default is 0.001.
        k_step: Step size for the shape parameter search.
        scale: Scale value (theta) for gamma distribution. Typically set at 1.
        target_probability: Threshold used in determining the shape parameter (k) of the
            gamma distribution. Probability that a value 3 standard deviations from the
            buffer zone is maxed-out.
        scale: Scale value (theta) for gamma distribution. Typically set at 1.

    Returns:
        The optimal shape parameter that makes the gamma CDF at the given standard
        deviation approximately equal to the target probability.
    """
    k_values = np.arange(k_max, k_min, -k_step)
    previous_probability = 1.0
    previous_k = 0.0
    result = 0.0

    for k in k_values:
        current_probability = stats.gamma.cdf(standard_deviation, a=k, scale=scale)

        if (
            previous_probability < target_probability
            and current_probability >= target_probability
        ):
            if abs(target_probability - previous_probability) > abs(
                current_probability - target_probability
            ):
                result = float(k)
            else:
                result = previous_k
            return result

        previous_probability = current_probability
        previous_k = float(k)

    return result


def _extrapolate_neighbors(
    marker: np.ndarray,
    confidence_threshold: float = 0.5,
    neighborhood_size: float = 0.05,
    sampling_rate: int = 100,
) -> pl.DataFrame:
    """Find neighborhoods around maxed-out regions for regression fitting.

    Args:
        marker: Array containing values that mark the probability that each sample is
            maxed-out. Values close to ±1 indicate high likelihood that the value is
            near the limit of the dynamic range, with the sign indicating if it's near
            the upper bound, or lower bound. Values close to zero indicate low
            likelihood of being near range limit.
        confidence_threshold: Threshold for what constitutes a significant change in
            marker values.
        neighborhood_size: Duration of neighborhood in seconds to use around each
            maxed-out region for regression fitting. Default is 0.05 (50ms).
        sampling_rate: The sampling rate of the data in Hz. Default is 100Hz as data
            should be interpolated to 100hz prior to this step.

    Returns:
        DataFrame containing the indices of maxed-out regions (start, end) and
        their expanded neighborhoods (left_neighborhood, right_neighborhood). A
        value of -1 indicates a boundary case where the region extends to the
        beginning or end of the data.
    """
    n_neighbor = int(sampling_rate * neighborhood_size)
    edges_indices = _extrapolate_edges(
        marker=marker,
        confidence_threshold=confidence_threshold,
        sampling_rate=sampling_rate,
    )
    marker_length = len(marker)

    expanded = edges_indices.with_columns(
        [
            (pl.col("start") - n_neighbor).alias("left_neighborhood"),
            (pl.col("end") + n_neighbor).alias("right_neighborhood"),
        ]
    )

    neighborhoods = expanded.with_columns(
        [
            pl.when(pl.col("start") == -1)
            .then(-1)
            .otherwise(pl.col("left_neighborhood").clip(0, marker_length - 1))
            .alias("left_neighborhood"),
            pl.when(pl.col("end") == -1)
            .then(-1)
            .otherwise(pl.col("right_neighborhood").clip(0, marker_length - 1))
            .alias("right_neighborhood"),
        ]
    )

    return neighborhoods


def _extrapolate_edges(
    marker: np.ndarray, confidence_threshold: float = 0.5, sampling_rate: int = 100
) -> pl.DataFrame:
    """Find edges of maxed out regions and identify them as hills(+) or valleys (-).

    Args:
        marker: Array containing values that mark the probability that each sample is
            maxed-out. Values close to ±1 indicate high likelihood that the value is
            near the limit of the dynamic range, with the sign indicating if it's near
            the upper bound, or lower bound. Values close to zero indicate low
            likelihood of being near range limit.
        confidence_threshold: Threshold for what constitutes a significant change in
            marker values.
        sampling_rate: Sampling rate of acceleration data, typically 100hz following
            interpolation.

    Returns:
        DataFrame containing indices for hills and valleys.
    """
    marker_diff_left = np.concatenate(([0], np.diff(marker)))
    marker_diff_right = np.concatenate((np.diff(marker), [0]))

    out_of_range_threshold = sampling_rate * 5

    positive_left_end = np.where(
        (marker_diff_left > confidence_threshold) & (marker > 0)
    )[0]
    positive_right_start = np.where(
        (marker_diff_right < -confidence_threshold) & (marker > 0)
    )[0]
    hills_df = _align_edges(
        marker_length=len(marker),
        left=positive_left_end,
        right=positive_right_start,
        out_of_range_threshold=out_of_range_threshold,
        sign="hill",
    )

    negative_left_end = np.where(
        (marker_diff_left < -confidence_threshold) & (marker < 0)
    )[0]
    negative_right_start = np.where(
        (marker_diff_right > confidence_threshold) & (marker < 0)
    )[0]
    valleys_df = _align_edges(
        marker_length=len(marker),
        left=negative_left_end,
        right=negative_right_start,
        out_of_range_threshold=out_of_range_threshold,
        sign="valley",
    )

    return pl.concat([hills_df, valleys_df], how="vertical")


def _align_edges(
    marker_length: int,
    left: np.ndarray,
    right: np.ndarray,
    out_of_range_threshold: int,
    sign: Literal["hill", "valley"],
) -> pl.DataFrame:
    """Aligns left and right edges of maxed-out regions and handles edge cases.

    Args:
        marker_length: Length of marker array.
        left: 1D vector of indices indicating the left edge of a potential maxed-out
            region.
        right: 1D vector of indices indicating the right edge of a potential maxed-out
            region.
        out_of_range_threshold: The number of samples that determines if an un matched
            edge extends beyond the data range, or if the edge is spurious. Typically 5
            seconds worth of data.
        sign: Indicates if maxed out region is a hill or valley.

    Returns:
        A DataFrame with the left and right edges of each maxed-out region.

    Raises:
        ValueError: If the difference in length between the left and right arrays is
        greater than 1.
    """
    if np.abs(len(left) - len(right)) >= 2:
        raise ValueError(
            f"Mismatch in {sign} edges. # left: {len(left)}, # right: {len(right)}"
        )

    if len(left) - len(right) == 1:
        if marker_length - left[-1] > out_of_range_threshold:
            right = np.concatenate((right, [-1]))
        elif len(left) == 1:
            left = np.array([])
        else:
            left = left[:-1]

    elif len(left) - len(right) == -1:
        if right[0] > out_of_range_threshold:
            left = np.concatenate(([-1], left))
        elif len(right) == 1:
            right = np.array([])
        else:
            right = right[1:]

    valid_indices = (left != -1) & (right != -1)
    if np.any((right[valid_indices] - left[valid_indices]) < 0) and (len(right) > 1):
        left = left[:-1]
        right = right[1:]

    maxed_out_areas = pl.DataFrame({"start": left, "end": right})

    return maxed_out_areas


def _extrapolate_fit(
    axis: np.ndarray,
    time_numeric: np.ndarray,
    marker: np.ndarray,
    neighbors: pl.DataFrame,
    smoothing: float = 0.6,
    neighborhood_size: float = 0.05,
    sampling_rate: int = 100,
) -> List[Tuple[float, float]]:
    """Extrapolate peak of maxed regions by fitting weighted splines on nearby points.

    Args:
        axis: Original acceleration data along one axis.
        time_numeric: Time series data given in epoch time (ns).
        marker: Array the same size as axis, containing values that mark the
            probability that each sample is maxed-out. Values close to ±1 indicate
            high likelihood that the value is near the limit of the dynamic range, with
            the sign indicating if it's near the upper bound, or lower bound. Values
            close to zero indicate low likelihood of being near range limit.
        neighbors: A polars DataFrame containing the start and end indices for the
            maxed out regions, as well as the neighborhoods to the left and right of
            regions.
        smoothing: Smoothing value for UniveriateSpline. Determines how closely the
            spline adheres to the data points. Higher values will give a smoother
            result.
        neighborhood_size: Duration of neighborhood in seconds to use around each
            maxed-out region for regression fitting. Default is 0.05 (50ms).
        sampling_rate: Sampling rate, default is 100Hz

    Returns:
        A List of Tuples representing the peaks of maxed out regions as
        (timestamp, value) pairs. Times are given in epoch time (ns).
    """
    extrapolate_points = []

    for row in neighbors.iter_rows(named=True):
        left_start, left_end = row["left_neighborhood"], row["start"]
        right_start, right_end = row["end"], row["right_neighborhood"]

        fitted_left = _fit_weighted(
            axis=axis,
            time_numeric=time_numeric,
            marker=marker,
            start=left_start,
            end=left_end,
            smoothing=smoothing,
            neighborhood_size=neighborhood_size,
            sampling_rate=sampling_rate,
        )
        fitted_right = _fit_weighted(
            axis=axis,
            time_numeric=time_numeric,
            marker=marker,
            start=right_start,
            end=right_end,
            smoothing=smoothing,
            neighborhood_size=neighborhood_size,
            sampling_rate=sampling_rate,
        )

        if fitted_left is None or fitted_right is None:
            continue

        middle_t = (time_numeric[left_end] + time_numeric[right_start]) / 2

        left_extrapolated = fitted_left(middle_t)
        right_extrapolated = fitted_right(middle_t)

        extrapolated_value = (
            np.array(left_extrapolated) + np.array(right_extrapolated)
        ) / 2

        extrapolate_points.append((middle_t, extrapolated_value))

    return extrapolate_points


def _fit_weighted(
    axis: np.ndarray,
    time_numeric: np.ndarray,
    marker: np.ndarray,
    start: int,
    end: int,
    smoothing: float = 0.6,
    neighborhood_size: float = 0.05,
    sampling_rate: int = 100,
) -> Optional[interpolate.UnivariateSpline]:
    """Fit weighted spline regression model to a section of the data.

    Args:
        axis: Original acceleration data along one axis.
        time_numeric: Time series data given in epoch time (ns).
        marker: Array the same size as axis, containing values that mark the
            probability that each sample is maxed-out. Values close to ±1 indicate
            high likelihood that the value is near the limit of the dynamic range, with
            the sign indicating if it's near the upper bound, or lower bound. Values
            close to zero indicate low likelihood of being near range limit.
        start: Index marking the beginning of the maxed out zone.
        end: Index marking end of maxed out zone.
        smoothing: Smoothing value for UniveriateSpline. Determines how closely the
            spline adheres to the data points. Higher values will give a smoother
            result. Default is 0.6 as optimized in the MIMS-unit paper.
        neighborhood_size: Duration of neighborhood in seconds to use around each
            maxed-out region for regression fitting. Default is 0.05 (50ms).
        sampling_rate: Sampling rate, default is 100Hz

    Returns:
        A UnivariateSpline function fitted to the data segment with weights based on
        the probability of not being maxed-out. Returns None if the input data is
        insufficient or invalid (-1).
    """
    n_over = int(sampling_rate * neighborhood_size)
    if n_over < 4:
        return None
    if start < 0 or end < 0:
        return None

    sub_t = time_numeric[start : end + 1]
    sub_value = axis[start : end + 1]
    weight = 1 - marker[start : end + 1]

    if len(sub_t) < 2:
        return None

    over_t = np.linspace(sub_t[0], sub_t[-1], n_over)
    over_value = np.interp(over_t, sub_t, sub_value)
    over_weight = np.interp(over_t, sub_t, weight)

    fitted = interpolate.UnivariateSpline(
        over_t, over_value, w=over_weight, s=smoothing
    )

    return fitted


def _extrapolate_interpolate(
    axis: np.ndarray,
    time_numeric: np.ndarray,
    marker: np.ndarray,
    points: list,
    confidence_threshold: float = 0.5,
) -> np.ndarray:
    """Interpolate the original signal with extrapolated points.

    Args:
        axis: Original acceleration data along one axis.
        time_numeric: Time series data given in epoch time (ns).
        marker: Array the same size as axis, containing values that mark the
            probability that each sample is maxed-out. Values close to ±1 indicate
            high likelihood that the value is near the limit of the dynamic range, with
            the sign indicating if it's near the upper bound, or lower bound. Values
            close to zero indicate low likelihood of being near range limit.
        points: List of tuples with extrapolated points (time, value). Times are given
            in epoch time (ns).
        confidence_threshold: Threshold for maxed-out identification.

    Returns:
        np.ndarray of axis data, with extrapolated values.
    """
    non_maxed_out_mask = np.abs(marker) < confidence_threshold
    num_non_maxed = np.sum(non_maxed_out_mask)

    if num_non_maxed / len(marker) < 0.3:
        times = time_numeric
        values = axis
    else:
        times = time_numeric[non_maxed_out_mask]
        values = axis[non_maxed_out_mask]

        if points and len(points) > 0:
            points_time = np.array([point[0] for point in points])
            points_value = np.array([point[1] for point in points])

            times = np.concatenate([times, points_time])
            values = np.concatenate([values, points_value])

            sort_idx = np.argsort(times)
            times = times[sort_idx]
            values = values[sort_idx]

    interp_function = interpolate.CubicSpline(times, values, bc_type="natural")
    interp_values = interp_function(time_numeric)

    return interp_values


def butterworth_filter(
    acceleration: models.Measurement,
    sampling_rate: int = 100,
    cutoffs: tuple[float, float] = (0.2, 5.0),
    order: int = 4,
) -> models.Measurement:
    """Apply butterworth IIR filter to acceleration data.

    Implements third portion of MIMS algorithm following interpolate, and extrapolation.

    Args:
        acceleration: Acceleration data to be filtered.
        sampling_rate: Sampling rate of acceleration data in Hz.
        cutoffs: Cutoff values for bandpass filter.
        order: Order of the filter, defaults to 4th order.


    Returns:
        Acceleration Measurement of filtered data.
    """
    normalized_cutoffs = [cutoff / (sampling_rate * 0.5) for cutoff in cutoffs]
    b, a = signal.butter(N=order, Wn=normalized_cutoffs, btype="bandpass")

    filtered_data = [
        signal.lfilter(b=b, a=a, x=column) for column in acceleration.measurements.T
    ]

    return models.Measurement(
        measurements=np.column_stack(filtered_data), time=acceleration.time
    )


def aggregate_mims(
    acceleration: models.Measurement,
    epoch: float = 60.0,
    sampling_rate: int = 100,
    *,
    rectify: bool = True,
    truncate: bool = True,
) -> models.Measurement:
    """Calculate the area under the curve(AUC), per epoch, per axis.

    When an epoch has less than 90% of the expected values (based on the sampling rate
    and epoch length), the AUC for that epoch is given as -1 for each axis. If rectify
    is True, any axis with values below -150 will have the AUC value for that axis
    be -1 for that epoch. Finally, following integration, any value greater than (16 *
    sampling_rate * epoch) will be set to -1.

    Args:
        acceleration: Acceleration data to be aggregated.
        epoch: The desired epoch length in seconds that data will be aggregated over.
        sampling_rate: The sampling rate of the accelerometer data in Hz.
        rectify: Specifies if data should be rectified before integration. If True any
            value below -150 will assign the value of that axis to -1 for that epoch.
            Additionally the absolute value of accelerometer data will be used for
            integration.
        truncate: Specifies if data <= 0.001 should be truncated to 0.

    Returns:
        A models.Measurement instance with the area under the curve values for each
        epoch.
    """
    if epoch <= 0:
        raise ValueError("Epoch length must be greater than 0.")

    epoch_ns = int(epoch * 1e9)
    acceleration_df = pl.DataFrame(
        {
            "time": acceleration.time,
            "x": acceleration.measurements[:, 0],
            "y": acceleration.measurements[:, 1],
            "z": acceleration.measurements[:, 2],
        }
    ).set_sorted("time")

    result = acceleration_df.group_by_dynamic(
        "time",
        every=f"{epoch_ns}ns",
    ).map_groups(
        lambda group: _aggregate_epoch(
            group=group,
            epoch=epoch,
            sampling_rate=sampling_rate,
            rectify=rectify,
        ),
        schema={
            "time": pl.Datetime("ns"),
            "x": pl.Float64,
            "y": pl.Float64,
            "z": pl.Float64,
        },
    )

    result = result.with_columns(pl.col("time").dt.truncate(f"{int(epoch)}s"))

    aggregated_measure = models.Measurement(
        measurements=result.select(["x", "y", "z"]).to_numpy(),
        time=result["time"].cast(pl.Datetime("ns")),
    )

    if truncate:
        aggregated_measure.measurements = np.where(
            aggregated_measure.measurements <= 0.001, 0, aggregated_measure.measurements
        )

    return aggregated_measure


def _aggregate_epoch(
    group: pl.DataFrame,
    epoch: float = 60.0,
    sampling_rate: int = 100,
    *,
    rectify: bool = True,
) -> pl.DataFrame:
    """Calculate the area under the curve(AUC), per epoch.

    Args:
        group: The epoch given by .map_groups()
        epoch: The desired epoch length in seconds that data will be aggregated over.
            Defaults to 1 minute as per MIMS-unit paper.
        sampling_rate: The sampling rate of the accelerometer data in Hz.
        rectify: Specifies if data should be rectified before integration. If True any
            value below -150 will assign the value of that axis to -1 for that epoch.
            Additionally the absolute value of accelerometer data will be used for
            integration.

    Returns:
        A polars DataFrame containing the XYZ AUC values for one epoch.
    """
    timestamps = group["time"].cast(pl.Int64).to_numpy()

    if len(timestamps) < 0.9 * sampling_rate * epoch:
        logger.debug("Not enough data to calculate AUC")
        return pl.DataFrame(
            {"time": [group["time"].min()], "x": [-1.0], "y": [-1.0], "z": [-1.0]}
        )

    values = group.select(["x", "y", "z"]).to_numpy()

    times_sec = timestamps / 1e9

    if rectify:
        area = integrate.trapezoid(y=np.abs(values), x=times_sec, axis=0)
        low_values = np.any(values <= -150, axis=0)
        if np.any(low_values):
            logger.warning("Values below -150 detected. Consider checking your data.")
        area = np.where(low_values, -1, area)
    else:
        area = integrate.trapezoid(y=values, x=times_sec, axis=0)

    max_value = 16 * sampling_rate * epoch
    area = np.where(np.logical_or(area >= max_value, area < 0), -1.0, area)

    return pl.DataFrame(
        {"time": [group["time"].min()], "x": [area[0]], "y": [area[1]], "z": [area[2]]}
    )


def combine_mims(
    acceleration: models.Measurement,
    combination_method: Literal["sum", "vector_magnitude"] = "sum",
) -> models.Measurement:
    """Combine MIMS values of xyz axis into one MIMS value.

    If any value in an epoch is -1, the 'combined_mims' value for that epoch will
    be set to -1 to flag it as an invalid epoch.

    Args:
        acceleration: An object containing per-axis MIMS measurements and their
            corresponding timestamps.
        combination_method: The method to combine MIMS values across axes. Defaults to
            "sum".

    Returns:
            A Measurement object with combined MIMS values and corresponding timestamps.
    """
    row_contains_negative = np.any(acceleration.measurements == -1, axis=1)

    if combination_method == "sum":
        row_sum = np.sum(acceleration.measurements, axis=1)
        combined_mims = np.where(row_contains_negative, -1, row_sum)
    elif combination_method == "vector_magnitude":
        row_norm = np.linalg.norm(acceleration.measurements, axis=1)
        combined_mims = np.where(row_contains_negative, -1, row_norm)
    else:
        raise ValueError(
            f"Invalid combination_method given:{combination_method}."
            "Must be 'sum' or 'vector_magnitude'. "
        )

    return models.Measurement(measurements=combined_mims, time=acceleration.time)



"""Custom exceptions for wristpy."""

from wristpy.core import config

logger = config.get_logger()


class LoggedException(Exception):
    """Base class that automatically logs messages."""

    def __init__(self, message: str) -> None:
        """Initialize a new instance of the LoggedException class.

        Args:
            message: The message to display.
        """
        logger.exception(message)
        super().__init__(message)


class SphereCriteriaError(LoggedException):
    """Data did not meet the sphere criteria."""

    pass


class CalibrationError(LoggedException):
    """Custom exception for failures in calibration."""

    pass


class NoMotionError(LoggedException):
    """No epochs with zero movement could be found in the data."""

    pass


class InvalidFileTypeError(LoggedException):
    """Wristpy did not expect this file extension."""

    pass



"""CLI for wristpy."""

import logging
import pathlib
from enum import Enum

import typer

from wristpy.core import config

logger = config.get_logger()
app = typer.Typer(
    help="Run the main Wristpy pipeline.",
    epilog="Please report issues at https://github.com/childmindresearch/wristpy.",
)


class OutputFileType(str, Enum):
    """Valid output file types for saving data."""

    csv = ".csv"
    parquet = ".parquet"


class Calibrator(str, Enum):
    """Setting a calibrator class for typer.

    This class is used to define the literal types that are allowed for
    calibration, and parsing the strings for the orchestrator.
    """

    none = "none"
    ggir = "ggir"
    gradient = "gradient"


class ActivityMetric(str, Enum):
    """Valid activity metrics for physical activity categorization."""

    enmo = "enmo"
    mad = "mad"
    ag_count = "ag_count"
    mims = "mims"


class NonwearAlgorithms(str, Enum):
    """Setting a nonwear algorithm class for typer.

    This class is used to define the literal types that are allowed for
    nonwear algorithms, and parsing the strings for the orchestrator.
    """

    ggir = "ggir"
    cta = "cta"
    detach = "detach"


def version_check(version: bool) -> None:
    """Print the current version of wristpy and exit."""
    if version:
        typer.echo(f"Wristpy version: {config.get_version()}")
        raise typer.Exit()


@app.command()
def main(
    input: pathlib.Path = typer.Argument(
        ..., help="Path to the input data.", exists=True
    ),
    output: pathlib.Path = typer.Option(
        None,
        "-o",
        "--output",
        help="Path where data will be saved. Supports .csv and .parquet formats.",
    ),
    output_filetype: OutputFileType = typer.Option(
        None,
        "-O",
        "--output-filetype",
        help="Format for save files when processing directories. "
        "Leave as None when processing single files.",
    ),
    calibrator: Calibrator = typer.Option(
        Calibrator.none,
        "-c",
        "--calibrator",
        help="Pick which calibrator to use. "
        "Must choose one of 'none', 'ggir', or 'gradient'.",
        case_sensitive=False,
    ),
    activity_metric: ActivityMetric = typer.Option(
        ActivityMetric.enmo,
        "-a",
        "--activity-metric",
        help="Metric used for physical activity categorization. "
        "Choose from 'enmo', 'mad', 'ag_count', or 'mims'.  ",
        case_sensitive=False,
    ),
    thresholds: tuple[float, float, float] = typer.Option(
        None,
        "-t",
        "--thresholds",
        help="Provide three thresholds for light, moderate, and vigorous activity. "
        "Exactly three values must be >= 0, given in ascending order, "
        "and separated by a space. (e.g. '-t 0.1 1.0 1.5').",
        min=0,
    ),
    nonwear_algorithm: list[NonwearAlgorithms] = typer.Option(
        [NonwearAlgorithms.ggir],
        "-n",
        "--nonwear-algorithm",
        help="Specify the non-wear detection algorithm(s) to use. "
        "Specify one or more of 'ggir', 'cta', 'detach'. "
        "(e.g. '-n ggir -n cta'). "
        "When multiple algorithms are specified, majority voting will be applied.",
    ),
    epoch_length: int = typer.Option(
        5,
        "-e",
        "--epoch-length",
        help="Specify the sampling rate in seconds for all metrics. "
        "Must be greater than or equal to 1.",
        min=1,
    ),
    verbosity: bool = typer.Option(
        False,
        "-v",
        "--verbosity",
        help="Determines the level of verbosity. Use -v for DEBUG. "
        "Defaults to INFO if not included.",
    ),
    version: bool = typer.Option(
        False,
        "-V",
        "--version",
        help="Print the current version of wristpy and exit.",
        is_eager=True,
        callback=version_check,
    ),
) -> None:
    """Run wristpy orchestrator with command line arguments."""
    from wristpy.core import orchestrator

    log_level = logging.INFO
    if verbosity:
        log_level = logging.DEBUG
    logger.setLevel(log_level)

    nonwear_algorithms = [algo.value for algo in nonwear_algorithm]
    calibrator_value = None if calibrator == Calibrator.none else calibrator.value

    logger.debug("Running wristpy. arguments given: %s", locals())
    orchestrator.run(
        input=input,
        output=output,
        calibrator=calibrator_value,
        activity_metric=activity_metric.value,
        thresholds=None if thresholds is None else thresholds,
        epoch_length=epoch_length,
        nonwear_algorithm=nonwear_algorithms,  # type: ignore[arg-type] # Covered by NonwearAlgorithm Enum class
        verbosity=log_level,
        output_filetype=output_filetype.value if output_filetype else None,  # type: ignore[arg-type] # Covered by OutputFileType Enum class
    )


if __name__ == "__main__":
    app()



"""This module contains helper functions for aggregated nonwear detection outputs."""

import datetime
from typing import Literal, Sequence, Union

import numpy as np
import polars as pl

from wristpy.core import computations, models
from wristpy.processing import metrics


def _majority_vote_non_wear(
    nonwear_measurements: Sequence[models.Measurement],
    temporal_resolution: float = 60.0,
) -> models.Measurement:
    """This function applies a majority vote on any number of nonwear Measurements.

    The _time_fix function is used to ensure that all nonwear Measurements have the
    same start and endpoints. Then each nonwear Measurement is resampled to the same
    temporal_resolution. A majority vote is taken at each time point to determine the
    new nonwear Measurement.
    In case of an even number of inputs, the majority is rounded up.
    If only one nonwear Measurement is provided, it is returned as is.

    Args:
        nonwear_measurements: A sequence (ex. List, Tuple, ...) of Measurement objects.
        temporal_resolution: The temporal resolution of the output, in seconds.
            Defaults to 60.0.

    Returns:
        A new Measurement instance with the combined nonwear detection,
        at a new temporal resolution.

    Raises:
        ValueError: If the number of nonwear measurements is 0.
    """
    num_measurements = len(nonwear_measurements)

    if num_measurements == 0:
        raise ValueError("At least one nonwear measurement is required.")

    if num_measurements == 1:
        return nonwear_measurements[0]

    majority_threshold = num_measurements / 2

    min_start_time = min(measurement.time[0] for measurement in nonwear_measurements)
    max_end_time = max(measurement.time[-1] for measurement in nonwear_measurements)

    first_resampled_measurement = computations.resample(
        _time_fix(nonwear_measurements[0], max_end_time, min_start_time),
        temporal_resolution,
    )
    measurement_sum = np.zeros_like(first_resampled_measurement.measurements)
    reference_time = first_resampled_measurement.time

    for measurement in nonwear_measurements:
        time_adjust_measurement = _time_fix(measurement, max_end_time, min_start_time)
        resampled_measurement = computations.resample(
            time_adjust_measurement, temporal_resolution
        )

        binary_nonwear = np.where(resampled_measurement.measurements >= 0.5, 1, 0)
        measurement_sum += binary_nonwear

    nonwear_value = np.where(measurement_sum > majority_threshold, 1, 0)

    return models.Measurement(measurements=nonwear_value, time=reference_time)


def _time_fix(
    nonwear: models.Measurement,
    end_time: datetime.datetime,
    start_time: datetime.datetime,
) -> models.Measurement:
    """Helper function to fix the time of the nonwear measurements.

    This function appends start/end points to the nonwear measurements based on
    previously computed reference start and end points.

    Args:
        nonwear: The nonwear measurement to adjust start/end time points.
        end_time: The maximum end time of the nonwear measurements.
        start_time: The minimum start time of the nonwear measurements.

    Returns:
        The nonwear measurement with the time fixed.
    """
    if nonwear.time[0] > start_time:
        nonwear.time = pl.concat(
            [pl.Series([start_time], dtype=pl.Datetime("ns")), nonwear.time]
        )
        nonwear.measurements = np.append(nonwear.measurements[0], nonwear.measurements)

    if nonwear.time[-1] < end_time:
        nonwear.time.append(pl.Series([end_time], dtype=pl.Datetime("ns")))
        nonwear.measurements = np.append(nonwear.measurements, nonwear.measurements[-1])
    return nonwear


def get_nonwear_measurements(
    calibrated_acceleration: models.Measurement,
    temperature: Union[models.Measurement, None] = None,
    non_wear_algorithms: Sequence[Literal["ggir", "cta", "detach"]] = ["ggir"],
) -> models.Measurement:
    """Non-wear measurement dispatcher.

    This function chooses which non-wear detection algorithm(s) to use based on the
    provided sequence of valid algorithm names.

    Args:
        calibrated_acceleration: The calibrated acceleration data
        temperature: Temperature data if available
        non_wear_algorithms: One or more algorithm names to use.


    Returns:
        A non-wear Measurement object.

    Raises:
        ValueError:
            If the CTA or DETACH algorithm is requested without temperature data.
            If an unknown algorithm is specified.

    """
    temp_dependent_algorithm = {"cta", "detach"}
    if (
        any(algorithm in temp_dependent_algorithm for algorithm in non_wear_algorithms)
        and temperature is None
    ):
        raise ValueError(
            "Temperature data is required for the CTA and DETACH nonwear algorithms."
        )

    non_wear_algorithm_functions = {
        "ggir": lambda: metrics.detect_nonwear(calibrated_acceleration),
        "cta": lambda: metrics.combined_temp_accel_detect_nonwear(
            acceleration=calibrated_acceleration,
            temperature=temperature,  # type: ignore[arg-type] #protected by if statement
        ),
        "detach": lambda: metrics.detach_nonwear(
            acceleration=calibrated_acceleration,
            temperature=temperature,  # type: ignore[arg-type] #protected by if statement
        ),
    }

    if any(
        algorithm not in non_wear_algorithm_functions
        for algorithm in non_wear_algorithms
    ):
        raise ValueError("An unknown algorithm was specified.")

    results = [
        non_wear_algorithm_functions[algorithm]() for algorithm in non_wear_algorithms
    ]

    if len(results) > 1:
        return _majority_vote_non_wear(results)
    return results[0]


def nonwear_array_cleanup(
    nonwear_array: models.Measurement,
    reference_measurement: models.Measurement,
    epoch_length: float = 5.0,
) -> models.Measurement:
    """This function is used to match the nonwear array to a reference Measurement.

    This function ensures that the nonwear array and reference Measurement times
    are synced up. This is accomplished by first resampling the nonwear array to
    the specified temporal resolution.
    It also ensures that the nonwear array is a binary array, where 1 indicates
    nonwear and 0 indicates wear.
    It then truncates the nonwear array to match the reference Measurement time points.

    Args:
        nonwear_array: The nonwear array to clean up.
        reference_measurement: The reference measurement to use for resampling.
        epoch_length: The temporal resolution of the output, in seconds.
            Defaults to 5.0.

    Returns:
        A new Measurement instance with the cleaned up nonwear detection,
        returned as a boolean (True == nonwear).
    """
    time_fix_nonwear = _time_fix(
        nonwear_array, reference_measurement.time[-1], reference_measurement.time[0]
    )
    resampled_nonwear = computations.resample(time_fix_nonwear, epoch_length)
    binary_nonwear = np.where(resampled_nonwear.measurements >= 0.5, 1, 0)

    ref_df = pl.DataFrame({"time": reference_measurement.time})

    nonwear_df = pl.DataFrame(
        {
            "time": resampled_nonwear.time,
            "idx": pl.arange(0, len(resampled_nonwear.time), eager=True),
        }
    )

    joined = ref_df.join(nonwear_df, on="time", how="inner")

    matched_indices = joined["idx"].to_numpy()
    matched_values = binary_nonwear[matched_indices]

    return models.Measurement(
        measurements=matched_values.astype(bool), time=joined["time"]
    )



"""The core submodule contains key classes and functions.

These classes and functions are used to organize the pipeline processing
and provide functions for processing the various metrics and output analysis.
"""



"""This is the preprocessing submodule.

This module contains the functionality necessary to preprocess the raw data.
This includes calibration, deriving the various metrics, and the sleep
and physical activity analysis tools.
"""



"""Module containing the output classes for writing data to files."""

import datetime
import json
import pathlib
from typing import Any, Dict, Optional

import polars as pl
import pydantic

from wristpy.core import config, exceptions, models

VALID_FILE_TYPES = (".csv", ".parquet")

logger = config.get_logger()


class OrchestratorResults(pydantic.BaseModel):
    """Dataclass containing results of orchestrator.run()."""

    physical_activity_metric: models.Measurement
    anglez: models.Measurement
    physical_activity_levels: models.Measurement
    nonwear_status: models.Measurement
    sleep_status: models.Measurement
    processing_params: Optional[Dict[str, Any]] = None

    def save_results(self, output: pathlib.Path) -> None:
        """Convert to polars and save the dataframe as a csv or parquet file.

        Args:
            output: The path and file name of the data to be saved. as either a csv or
                parquet files.

        """
        logger.debug("Saving results.")
        self.validate_output(output=output)
        output.parent.mkdir(parents=True, exist_ok=True)

        results_dataframe = pl.DataFrame(
            {"time": self.physical_activity_metric.time}
            | {
                name: value.measurements
                for name, value in self
                if name not in "processing_params"
            }
        )

        if output.suffix == ".csv":
            results_dataframe.write_csv(output, separator=",")
        elif output.suffix == ".parquet":
            results_dataframe.write_parquet(output)

        logger.info("Results saved in: %s", output)

        if self.processing_params:
            self.save_config_as_json(output)

    def save_config_as_json(self, output_path: pathlib.Path) -> None:
        """Save processing parameters as a JSON configuration file.

        Args:
            output_path: Path where the data file was saved. The JSON file will use
                the same name but with .json extension.
        """
        if not self.processing_params:
            logger.warning("No processing parameters to save as JSON")
            return

        wristpy_version = config.get_version()

        config_data = {
            "processing_time": datetime.datetime.now().isoformat(timespec="seconds"),
            "wristpy_version": wristpy_version,
            "processing_parameters": self.processing_params,
        }

        config_path = output_path.with_suffix(".json")

        with open(config_path, "w") as f:
            json.dump(config_data, f, indent=4)

        logger.debug("Configuration saved in: %s", config_path)

    @classmethod
    def validate_output(cls, output: pathlib.Path) -> None:
        """Validates that the output path exists and is a valid format.

        Args:
            output: the name of the file to be saved, and the directory it will
                be saved in. Must be a .csv or .parquet file.

        Raises:
            InvalidFileTypeError:If the output file path ends with any extension other
                    than csv or parquet.
        """
        if output.suffix not in VALID_FILE_TYPES:
            raise exceptions.InvalidFileTypeError(
                f"The extension: {output.suffix} is not supported."
                "Please save the file as .csv or .parquet",
            )



"""Configuration module for wristpy."""

import logging
from importlib import metadata


def get_version() -> str:
    """Return wristpy version."""
    try:
        return metadata.version("wristpy")
    except metadata.PackageNotFoundError:
        return "Version unknown"


def get_logger() -> logging.Logger:
    """Gets the wristpy logger."""
    logger = logging.getLogger("wristpy")
    if logger.handlers:
        return logger

    logger.setLevel(logging.INFO)
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(filename)s:%(lineno)s - %(funcName)s - %(message)s",  # noqa: E501
    )

    handler = logging.StreamHandler()
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    return logger



"""Function to read accelerometer data from a file."""

import pathlib
from typing import Literal, Union

import actfast
import numpy as np
import polars as pl

from wristpy.core import models


def read_watch_data(file_name: Union[pathlib.Path, str]) -> models.WatchData:
    """Read watch data from a file.

    Currently supported watch types are Actigraph .gt3x and GeneActiv .bin.
    Assigns the idle_sleep_mode_flag to false unless the watchtype is .gt3x and
    sleep_mode is enabled (based on watch metadata).

    Args:
        file_name: The filename to read the watch data from.

    Returns:
        WatchData class

    Raises: IOError if the file extension is not supported or doesn't exist.
    """
    try:
        data = actfast.read(file_name)
    except Exception as e:
        raise IOError(f"Error reading file: {e}. File type is unsupported.") from e

    measurements: dict[str, models.Measurement] = {}

    for timeseries in data["timeseries"].values():
        time = unix_epoch_time_to_polars_datetime(timeseries["datetime"])
        for sensor_name, sensor_values in timeseries.items():
            measurements[sensor_name] = models.Measurement(
                measurements=sensor_values, time=time
            )

    file_type = pathlib.Path(file_name).suffix
    idle_sleep_mode_flag = False
    if file_type == ".gt3x":
        idle_sleep_mode_flag = (
            data["metadata"]["device_feature_enabled"]["sleep_mode"].lower() == "true"
        )

    dynamic_range = _extract_dynamic_range(
        metadata=data["metadata"],
        file_type=file_type,  # type: ignore[arg-type]
    )

    return models.WatchData(
        acceleration=measurements["acceleration"],
        lux=measurements.get("light"),
        battery=measurements.get("battery_voltage"),
        capsense=measurements.get("capsense"),
        temperature=measurements.get("temperature"),
        idle_sleep_mode_flag=idle_sleep_mode_flag,
        dynamic_range=dynamic_range,
    )


def _extract_dynamic_range(
    metadata: dict, file_type: Literal[".gt3x", ".bin"]
) -> tuple[float, float]:
    """Extract the dynamic range from metadata.

    Args:
        metadata: Metadata subdictionary where accelerometer range values can be found.
        file_type: Accelerometer data file type. Supports .gt3x and .bin.

    Returns:
        A tuple containing the accelerometer range.

    Raises:
        ValueError: If file type is not supported.
    """
    if file_type == ".gt3x":
        return (
            float(metadata.get("info", {}).get("Acceleration Min")),
            float(metadata.get("info", {}).get("Acceleration Max")),
        )
    elif file_type == ".bin":
        range_str = (
            metadata.get("Device Capabilities", {})
            .get("Accelerometer Range")
            .strip()
            .split(" to ")
        )
        return (float(range_str[0]), float(range_str[1]))

    raise ValueError(f"Unsupported file type given: {file_type}")


def unix_epoch_time_to_polars_datetime(
    time: np.ndarray, units: Literal["ns", "us", "ms", "s", "d"] = "ns"
) -> pl.Series:
    """Convert unix epoch time to polars Series of datetime.

    Args:
        time: The unix epoch timestamps to convert.
        units: The units to convert the time to ('s', 'ms', 'us', or 'ns'). Default
            value is 'ns'.
    """
    time_series = pl.Series(time)
    return pl.from_epoch(time_series, time_unit=units).alias("time")



"""This module will contain functions to compute statistics on the sensor data."""

from typing import Literal

import polars as pl

from wristpy.core import models


def _moving(
    measurement: models.Measurement,
    epoch_length: float,
    aggregation: Literal["mean", "std", "median"],
    *,
    centered: bool = False,
    continuous: bool = False,
) -> models.Measurement:
    """Internal handler of rolling window functions.

    Args:
        measurement: The measurement to apply a rolling function to.
        epoch_length: Length of the window in seconds.
        aggregation: Name of the function to apply, either 'mean', 'std', or 'median'.
        centered: If true, centers the window. Defaults to False.
        continuous: If true, applies the window to every measurement. If false,
            groups measurements into chunks of epoch_length. Defaults to False.

    Returns:
        The measurement with the rolling function applied to it.
    """
    if epoch_length <= 0:
        raise ValueError("Epoch length must be greater than 0")

    window_size = f"{int(epoch_length * 1e9)}ns"
    if centered:
        offset = f"{-int((epoch_length // 2 + 1) * 1e9)}ns"
    else:
        offset = "0ns"

    if continuous:
        aggregator = getattr(pl, aggregation)("*").rolling(
            index_column="time", period=window_size, offset=offset
        )
        aggregated_df = (
            measurement.lazy_frame()
            .select([aggregator])
            .with_columns(time=measurement.time)
        )
    else:
        aggregator = getattr(pl.exclude(["time"]).drop_nans(), aggregation)
        aggregated_df = (
            measurement.lazy_frame()
            .group_by_dynamic("time", every=window_size, offset=offset)
            .agg(aggregator())
        )

    return models.Measurement.from_data_frame(aggregated_df.collect())


def moving_mean(
    array: models.Measurement, epoch_length: float = 5
) -> models.Measurement:
    """Calculate the moving mean of the sensor data in array.

    Args:
        array: The Measurement object with the sensor data we want to take the mean of
        epoch_length: The length, in seconds, of the window.

    Returns:
        The moving mean of the array in a new Measurement instance.

    Raises:
        ValueError: If the epoch length is not an integer or is less than 1.
    """
    return _moving(array, epoch_length, "mean")


def moving_std(
    array: models.Measurement, epoch_length: float = 5
) -> models.Measurement:
    """Calculate the moving standard deviation (std) of the sensor data in array.

    Args:
        array: The Measurement object with the sensor data we want to take the std of
        epoch_length: The length, in seconds, of the window.

    Returns:
        The moving std of the array in a new Measurement instance.

    Raises:
        ValueError: If the epoch length is less than 1.
    """
    return _moving(array, epoch_length, "std")


def moving_median(array: models.Measurement, epoch_length: float) -> models.Measurement:
    """Applies moving median to acceleration data.

    Step size for the window is hard-coded to 1 sample.

    Args:
        array: The Measurement object with the sensor data we want to take the median
            of.
        epoch_length: Size of the moving median window. Window is centered.
            Measured in seconds.

    Returns:
        Measurement object with rolling median applied to the measurement data. The
        measurements data will retain its shape, and the time data will be returned
        unaltered.
    """
    return _moving(array, epoch_length, "median", centered=True, continuous=True)


def resample(measurement: models.Measurement, delta_t: float) -> models.Measurement:
    """Resamples a measurement to a different timescale.

    Args:
        measurement: The measurement to resample.
        delta_t: The new time step, in seconds. This will be rounded to the nearest
            nanosecond.

    Returns:
        The resampled measurement.

    Raises:
        ValueError: Raised for zero or negative delta_t.
    """
    if delta_t <= 0:
        msg = "delta_t must be positive."
        raise ValueError(msg)

    time_delta_median = (measurement.time[1:] - measurement.time[:-1]).median()

    n_nanoseconds_in_second = 1_000_000_000
    current_delta_t = time_delta_median.total_seconds() * n_nanoseconds_in_second  # type: ignore[union-attr] #Guarded by Measurement validation for .time attribute
    requested_delta_t = round(delta_t * n_nanoseconds_in_second)

    measurement_df = (
        pl.from_numpy(measurement.measurements)
        .with_columns(time=measurement.time)
        .set_sorted("time")
    )

    if current_delta_t >= requested_delta_t:
        resampled_df = (
            (
                measurement_df.group_by_dynamic(
                    "time", every=f"{requested_delta_t}ns", start_by="datapoint"
                )
            )
            .agg(pl.exclude("time").mean())
            .upsample(
                time_column="time", every=f"{requested_delta_t}ns", maintain_order=True
            )
            .interpolate()
            .fill_null(strategy="forward")
        )

    else:
        resampled_df = measurement_df.group_by_dynamic(
            "time", every=f"{requested_delta_t}ns", start_by="datapoint"
        ).agg(pl.exclude("time").mean())

    new_measurement = (
        resampled_df.drop("time").to_numpy().reshape((len(resampled_df), -1)).squeeze()
    )
    return models.Measurement(
        measurements=new_measurement,
        time=resampled_df["time"],
    )



"""Calculate sleep onset and wake up times."""

import abc
import datetime
from dataclasses import dataclass
from typing import List, Tuple, Union

import numpy as np
import polars as pl

from wristpy.core import computations, config, models

logger = config.get_logger()


@dataclass
class SleepWindow:
    """Dataclass to store sleep window information.

    Attributes:
        onset: the predicted start time of the sleep window.
        wakeup: the predicted end time of the sleep window.
    """

    onset: Union[datetime.datetime, List]
    wakeup: Union[datetime.datetime, List]


class AbstractSleepDetector(abc.ABC):
    """Abstract class defining the interface for sleep detection algorithms."""

    @abc.abstractmethod
    def __init__(self, anglez: models.Measurement) -> None:
        """Initialization function for the sleep detection algorithm.

        Must contain the anglez data as an input.
        """
        pass

    @abc.abstractmethod
    def run_sleep_detection(self) -> List[SleepWindow]:
        """Sleep Detector must contain a run_sleep_detection function.

        The function must return a list of SleepWindow objects.
        """
        pass


class GgirSleepDetection(AbstractSleepDetector):
    """Sleep Detection algorithm based on the GGIR method.

    This class implements the GGIR method for sleep detection. The method uses the
    angle-z data to find periods of sleep. It returns a list of SleepWindow objects.

    Attributes:
        anglez (models.Measurement): The angle-z data, calculated from the calibrated
            accelerometer data.
    """

    def __init__(
        self,
        anglez: models.Measurement,
    ) -> None:
        """Initialize the SleepDetection class with default attributes.

        Args:
            anglez: the raw anglez data, calculated from calibrated acceleration.
        """
        self.anglez = anglez

    def run_sleep_detection(self) -> List[SleepWindow]:
        """Run the GGIR sleep detection.

        This algorithm uses the angle-z data to first find potential sleep periods
        (SPT windows) and then find sustained inactivity bouts (SIB periods). The
        sleep onset and wake up times are then calculated based on the overlap between
        the SPT windows and SIB periods.

        Returns:
            A list of SleepWindow instances, each instance contains a sleep onset/wakeup
            time pair.
        """
        logger.debug("Beginning sleep detection.")
        spt_window = self._spt_window(self.anglez)
        sib_periods = self._calculate_sib_periods(self.anglez)
        spt_window_periods = _find_periods(spt_window)
        sib_window_periods = _find_periods(sib_periods)
        sleep_onset_wakeup = self._find_onset_wakeup_times(
            spt_window_periods, sib_window_periods
        )
        logger.debug(
            "Sleep detection complete. Windows detected: %s", len(sleep_onset_wakeup)
        )
        return sleep_onset_wakeup

    def _spt_window(
        self, anglez_data: models.Measurement, threshold: float = 0.2
    ) -> models.Measurement:
        """Implement Heuristic distribution of z angle (HDCZA) for SPT window detection.

        This function finds the absolute difference of the anglez data over 5s windows.
        We find the 5-minute rolling median of that difference.
        Next, we find when that 5-minute median is above a specified threshold, taken as
        the new default value from the GGIR implementation of the HDCZA algorithm. This
        represents non-sleep candidates. The logical not of this is the sleep candidate.
        We then find long blocks (30 minutes) when the threshold criteria is met.
        Any gaps in SPT windows that are less than 60 minutes are filled.

        Args:
            anglez_data: the raw anglez data, calculated from calibrated acceleration.
            threshold: the threshold for the distribution of z angle, chosen as
                0.2 based on new GGIR default.

        Returns:
            A Measurement instance with values set to 1 indicating identified
            SPT windows and corresponding time stamps.

        References:
            van Hees, V.T., et al. Estimating sleep parameters using an
                accelerometer without sleep diary. Sci Rep 8, 12975 (2018).
                https://doi.org/10.1038/s41598-018-31266-z.
        """
        logger.debug("Finding spt windows, Threshold: %s", threshold)
        long_epoch_median = 300
        long_block = 360
        short_block_gap = 720

        anglez_abs_diff = self._compute_abs_diff_mean_anglez(anglez_data)
        anglez_median_long_epoch = computations.moving_median(
            anglez_abs_diff, long_epoch_median
        )
        non_sleep_candidates = (
            anglez_median_long_epoch.measurements >= threshold
        ).flatten()

        sleep_candidates = np.logical_not(
            _fill_false_blocks(non_sleep_candidates, long_block)
        )

        sleep_idx_array_filled = _fill_false_blocks(sleep_candidates, short_block_gap)
        return models.Measurement(
            measurements=sleep_idx_array_filled, time=anglez_median_long_epoch.time
        )

    def _calculate_sib_periods(
        self, anglez_data: models.Measurement, threshold_degrees: int = 5
    ) -> models.Measurement:
        """Find the sustained inactivity bouts.

        This function finds the absolute difference of the anglez data over 5s windows.
        We then find the 5-minute windows where all the differences are below a
        threshold (defaults to 5 degrees).

        Args:
            anglez_data: the raw anglez data, calculated from calibrated acceleration.
            threshold_degrees: the threshold, in degrees, for inactivity.

        Returns:
            A Measurement instance with values set to 1 indicating identified SIB
            windows, and corresponding time stamps.

        References:
            van Hees, V. T. et al. A Novel, Open Access Method to Assess Sleep
                Duration Using a Wrist-Worn Accelerometer. PLoS One 10, e0142533 (2015).
                https://doi.org/10.1371/journal.pone.0142533.
        """
        logger.debug("Calculating SIB period threshold: %s degrees", threshold_degrees)
        anglez_abs_diff = self._compute_abs_diff_mean_anglez(anglez_data)

        anglez_pl_df = pl.DataFrame(
            {"time": anglez_abs_diff.time, "angz_diff": anglez_abs_diff.measurements}
        )

        anglez_pl_df = anglez_pl_df.with_columns(pl.col("time").set_sorted())
        anglez_grouped_by_window_length = anglez_pl_df.group_by_dynamic(
            index_column="time", every="5m"
        ).agg([pl.all().exclude(["time"])])
        flag = anglez_grouped_by_window_length["angz_diff"].map_elements(
            lambda grouped_list_data: all(
                list_data < threshold_degrees for list_data in grouped_list_data
            ),
            return_dtype=pl.Boolean,
        )

        return models.Measurement(
            measurements=(flag.to_numpy()).astype(int),
            time=anglez_grouped_by_window_length["time"],
        )

    def _find_onset_wakeup_times(
        self,
        spt_periods: List[Tuple[datetime.datetime, datetime.datetime]],
        sib_periods: List[Tuple[datetime.datetime, datetime.datetime]],
    ) -> List[SleepWindow]:
        """Find the sleep onset and wake up times.

        This function is implemented as follows:
        First, we use the spt_window as a guider for potential sleep.
        Then we find overlapping sustained inactivity bouts (sib_periods == 1).
        Sleep onset is defined as the start of the first sib_period that overlaps
        with a specific spt_window.
        Sleep wakeup is the end of the last SIB that overlaps with a spt window.

        Args:
            spt_periods: the sleep period guider windows, this is computed from the
                _find_periods method, it is a list of tuples of start and end times
                of the periods.
            sib_periods: the sustained inactivity bouts,  this is computed from the
                _find_periods method, it is a list of tuples of start and end times
                of the periods.

        Returns:
            A SleepWindow instance with sleep onset and wake up times.
            If there is no overlap between spt_windows and sib_periods,
            the onset and wakeup lists will be empty.
        """
        sleep_windows = []
        for sleep_guide in spt_periods:
            min_onset = None
            max_wakeup = None
            for inactivity_bout in sib_periods:
                if (
                    inactivity_bout[0] <= sleep_guide[1]
                    and inactivity_bout[1] >= sleep_guide[0]
                ):
                    if min_onset is None or inactivity_bout[0] < min_onset:
                        min_onset = inactivity_bout[0]
                    if max_wakeup is None or inactivity_bout[1] > max_wakeup:
                        max_wakeup = inactivity_bout[1]
            if min_onset is not None and max_wakeup is not None:
                sleep_windows.append(SleepWindow(onset=min_onset, wakeup=max_wakeup))
        logger.debug("Sleep windows found: %s", len(sleep_windows))
        return sleep_windows

    def _compute_abs_diff_mean_anglez(
        self, anglez_data: models.Measurement, window_size_seconds: int = 5
    ) -> models.Measurement:
        """Helper function to compute the absolute difference of averaged anglez data.

        Args:
            anglez_data: the raw anglez data.
            window_size_seconds: the window size in seconds to average the anglez data.

        Returns:
            A Measurement instance with the absolute difference of the anglez data.
            Note that the length of the returned Measurement instance will be one
            less than the input anglez_data, this is because np.diff returns an array
            that is diff_size shorter than the input, where diff_size is the size of
            the difference step.
        """
        anglez_epoch1 = computations.moving_mean(anglez_data, window_size_seconds)
        absolute_diff = np.abs(np.diff(anglez_epoch1.measurements))

        return models.Measurement(
            measurements=absolute_diff, time=anglez_epoch1.time[1:]
        )


def _find_periods(
    window_measurement: models.Measurement,
) -> List[Tuple[datetime.datetime, datetime.datetime]]:
    """Find periods where window_measurement is equal to 1.

    This is a helper function to return the periods in the format of
    List [start_of_period, end_of_period], it is used in the
    GGIRSleepDetection class and when removing non-wear periods from sleep windows.

    Args:
        window_measurement: the Measurement instance, intended to be
            either the spt_window or sib_period.

    Returns:
        A list of tuples, where each tuple contains the start and end times of
        a period. For isolated ones the function returns the same start
        and end time. The list is sorted by time.
    """
    logger.debug("Finding periods in window measurement.")
    edge_detection = np.convolve([1, 3, 1], window_measurement.measurements, "same")
    single_one = np.nonzero(edge_detection == 3)[0]

    single_periods = [
        (window_measurement.time.item(int(idx)), window_measurement.time.item(int(idx)))
        for idx in single_one
    ]

    blocked_one_edge = np.nonzero(edge_detection == 4)[0]
    block_pairs = np.reshape(blocked_one_edge, (-1, 2))
    block_periods = [
        (window_measurement.time.item(idx[0]), window_measurement.time.item(idx[1]))
        for idx in block_pairs
    ]
    all_periods = single_periods + block_periods
    all_periods.sort()

    logger.debug("Found %s periods.", len(all_periods))
    return all_periods


def _fill_false_blocks(boolean_array: np.ndarray, gap_block: int) -> np.ndarray:
    """Helper function to fill gaps in any window that are less than gap_blocks.

    We find the first non-zero in the boolean_array, if there are none ,
    we return the initial array.
    We then iterate over the array and count every zero between ones
    (skipping the leading zeros),
    if that value is less than the gap_block, we fill in with ones.

    Args:
        boolean_array: A generic boolean array, typically the SPT window.
        gap_block: the length of the gap that needs to be filled.

    Returns:
        A booelan numpy array where true, typically, indicates the SPT windows.
    """
    n_zeros = 0
    first_one_idx = next(
        (index for index, value in enumerate(boolean_array) if value), None
    )
    if first_one_idx is None:
        return boolean_array

    for sleep_array_idx in range(first_one_idx, len(boolean_array)):
        sleep_value = boolean_array[sleep_array_idx]
        if not sleep_value:
            n_zeros += 1
            continue

        if n_zeros < gap_block:
            boolean_array[sleep_array_idx - n_zeros : sleep_array_idx] = True
        n_zeros = 0

    return boolean_array


def compute_physical_activty_categories(
    activity_metric_epoch1: models.Measurement,
    thresholds: Tuple[float, float, float] = (0.0563, 0.1916, 0.6958),
) -> models.Measurement:
    """Compute the physical activity categories based on the specific activity metric.

    This function uses the activity_metric_epoch1 data (5s aggregated data) to compute
    four physical activity levels: inactive, light, moderate, and vigorous.

    Default values are specifically for ENMO data, taken from the Hildebrand 2014 study,
    and is best suited for children aged 7 - 11 years.

    Args:
        activity_metric_epoch1: The specific activity metric epoch1 data, as physical
            activity data should be computed on aggregated data.
        thresholds: The threshold values for the physical activity categories.
            The default values are:
            (light_threshold, moderate_threshold, vigorous_threshold).

    Returns:
        A Measurement instance with the physical activity categories. Categories are
        Inactive, Light, Moderate, Vigorous. The temporal resolution is the same as
        activity_metric_epoch1.

    Raises:
        ValueError: If the threshold values are not poisitive, unique and  in ascending
        order.

    References:
        Hildebrand, M., et al. Age group comparability of raw accelerometer output
            from wrist- and hip-worn monitors. Medicine and Science in
            Sports and Exercise, 46(9), 1816-1824 (2014).
            https://doi.org/10.1249/mss.0000000000000289.
    """
    logger.debug("Computing physical activity levels, thresholds: %s", thresholds)
    if not (0 <= thresholds[0] < thresholds[1] < thresholds[2]):
        message = "Thresholds must be positive, unique, and given in ascending order."
        logger.error(message)
        raise ValueError(message)

    activity_levels = np.select(
        [
            activity_metric_epoch1.measurements <= thresholds[0],
            (thresholds[0] < activity_metric_epoch1.measurements)
            & (activity_metric_epoch1.measurements <= thresholds[1]),
            (thresholds[1] < activity_metric_epoch1.measurements)
            & (activity_metric_epoch1.measurements <= thresholds[2]),
            thresholds[2] < activity_metric_epoch1.measurements,
        ],
        ["inactive", "light", "moderate", "vigorous"],
        default="unknown",
    )
    return models.Measurement(
        measurements=activity_levels, time=activity_metric_epoch1.time
    )


def sleep_cleanup(
    sleep_windows: List[SleepWindow], nonwear_measurement: models.Measurement
) -> models.Measurement:
    """This function will filter the sleep windows based on the nonwear measurement.

    The SleepWindows are first converted to a Measurement object with the same
    timestamps as the reference measurement. Then any overlap with nonwear
    is removed, and finally any blocks of sleep that are less than 15 minutes
    long are removed.

    Args:
        sleep_windows: List of the sleep windows (Onset/Wake pairs).
        nonwear_measurement: The nonwear measurement data used for reference time
            stamps and to remove overlaps with periods of sleep.

    Returns:
        A Measurement instance with the cleaned sleep data.
    """
    logger.debug("Starting the sleep Window cleanup.")
    temporal_resolution = nonwear_measurement.time[1] - nonwear_measurement.time[0]
    num_samples_15min = int(15 * 60 / temporal_resolution.total_seconds())

    sleep = _sleep_windows_as_measurement(nonwear_measurement.time, sleep_windows)

    filtered_sleep = np.logical_and(
        sleep.measurements,
        np.logical_not(nonwear_measurement.measurements.astype(bool)),
    )
    cleaned_sleep = np.logical_not(
        _fill_false_blocks(np.logical_not(filtered_sleep), num_samples_15min)
    )

    return models.Measurement(time=sleep.time, measurements=cleaned_sleep)


def _sleep_windows_as_measurement(
    ref_measurement_time: pl.Series, sleep_windows: List[SleepWindow]
) -> models.Measurement:
    """Helper function to convert list of sleep windows to a Measurement instance.

    The temporal resolution of the output Measurement instance matches the
    reference measurement.

    Args:
        ref_measurement_time: Reference polars Series with time stamps from a
            reference Measurement.
        sleep_windows: The list of sleep windows, where the entries are
            instances of the SleepWindow class.

    Returns:
        A new Measurement instance with the sleep values, as a booleans.
    """
    logger.debug("Converting sleep windows to measurement.")

    sleep_value = np.zeros(len(ref_measurement_time), dtype=bool)

    for sw in sleep_windows:
        if sw.onset is not None and sw.wakeup is not None:
            time_mask = (ref_measurement_time >= sw.onset) & (
                ref_measurement_time <= sw.wakeup
            )
            sleep_value[time_mask] = True

    return models.Measurement(time=ref_measurement_time, measurements=sleep_value)



"""Calculate base metrics, anglez and enmo."""

from typing import Literal, Tuple

import numpy as np
import polars as pl
from scipy import interpolate, signal
from skdh.preprocessing import wear_detection

from wristpy.core import computations, config, models
from wristpy.io.readers import readers
from wristpy.processing import mims

logger = config.get_logger()


def euclidean_norm_minus_one(
    acceleration: models.Measurement, epoch_length: float = 5.0
) -> models.Measurement:
    """Compute ENMO, the Euclidean Norm Minus One (1 standard gravity unit).

    Negative values of ENMO are set to zero because ENMO is meant as a measure of
    physival activity. Negative values steming from imperfect calibration or noise from
    the device would have no meaningful interpretation in this context and would be
    detrimental to the intended analysis.

    The data can be downsampled to a temporal resolution of `epoch_length` seconds.

    Args:
        acceleration: the three-dimensional accelerometer data. A Measurement object,
        it will have two attributes. 1) measurements, containing the three-dimensional
        accelerometer data in an np.array and 2) time, a pl.Series containing
        datetime.datetime objects.
        epoch_length: The temporal resolution of the downsampled enmo, in seconds.
            Must be greater than 0. Defaults to 5.0s.

    Returns:
        A Measurement object containing the calculated ENMO values and the
        associated time stamps taken from the input.
    """
    enmo = np.linalg.norm(acceleration.measurements, axis=1) - 1

    enmo = np.maximum(enmo, 0)

    return computations.moving_mean(
        models.Measurement(measurements=enmo, time=acceleration.time), epoch_length
    )


def angle_relative_to_horizontal(
    acceleration: models.Measurement, epoch_length: float = 5.0
) -> models.Measurement:
    """Calculate the angle of the acceleration vector relative to the horizontal plane.

     The data can be downsampled to a temporal resolution of `epoch_length` seconds.

    Args:
        acceleration: the three-dimensional accelerometer data. A Measurement object,
        it will have two attributes. 1) measurements, containing the three-dimensional
        accelerometer data in a np.array and 2) time, a pl.Series containing
        datetime.datetime objects.
        epoch_length: The temporal resolution of the downsampled anglez, in seconds.
            Must be greater than 0. Defaults to 5.0s.

    Returns:
        A Measurement instance containing the values of the angle relative to the
        horizontal plane and the associated timestamps taken from the input unaltered.
        The angle is measured in degrees.
    """
    xy_projection_magnitute = np.linalg.norm(acceleration.measurements[:, 0:2], axis=1)

    angle_radians = np.arctan(acceleration.measurements[:, 2] / xy_projection_magnitute)

    angle_degrees = np.degrees(angle_radians)

    return computations.moving_mean(
        models.Measurement(measurements=angle_degrees, time=acceleration.time),
        epoch_length,
    )


def mean_amplitude_deviation(
    acceleration: models.Measurement, epoch_length: float = 5.0
) -> models.Measurement:
    """Calculate the mean amplitude deviation of the acceleration data.

    An alternative to ENMO to quantify the intensity of physical activity.
    It is calculated as the mean of the absolute difference between acceleration values
    and the mean of those acceleration values in a 5s window.

    Args:
        acceleration: the calibrated acceleration data.
        epoch_length: The length of the window in seconds.

    Returns:
        A new Measurement object containing the MAD values.

    References:
        Vähä-Ypyä H, Vasankari T, Husu P, Suni J, Sievänen H. A universal, accurate
        intensity-based classification of different physical activities using raw data
        of accelerometer. Clin Physiol Funct Imaging. 2015 Jan;35(1):64-70.
        doi: 10.1111/cpf.12127. Epub 2014 Jan 7. PMID: 24393233.
    """
    acceleration_magnitude = np.linalg.norm(acceleration.measurements, axis=1)

    mad_lf = pl.LazyFrame(
        {"time": acceleration.time, "acceleration_magnitude": acceleration_magnitude}
    ).set_sorted("time")

    mad_df = (
        mad_lf.group_by_dynamic(index_column="time", every=f"{int(epoch_length*1e9)}ns")
        .agg(
            [
                (
                    pl.col("acceleration_magnitude")
                    - pl.col("acceleration_magnitude").mean()
                )
                .abs()
                .mean()
                .alias("mean_amplitude_deviation")
            ]
        )
        .collect()
    )

    return models.Measurement.from_data_frame(mad_df)


def actigraph_activity_counts(
    acceleration: models.Measurement, epoch_length: float = 5.0
) -> models.Measurement:
    """Compute Actigraph acitivty counts.

    This function computes the Actigraph activity counts based on [1].
    The acceleration data is downsample to 30Hz, bandpass filtered, scaled,
    and then thresholded. The counts are then summed along each axis for
    the chosen epoch length.

    Args:
        acceleration: The calibrated acceleration data.
        epoch_length: The length of the epoch in seconds, defaults to 60s.

    Returns:
        The activity counts as a Measurement object.

    References:
        [1] A. Neishabouri et al., “Quantification of acceleration as activity counts
        in ActiGraph wearable,” Sci Rep, vol. 12, no. 1, Art. no. 1, Jul. 2022,
        doi: 10.1038/s41598-022-16003-x.
    """
    logger.debug("Running activity count physical activity metric.")

    epoch_length_nanoseconds = round(epoch_length * 1e9)

    # input and output coefficients for the bandpass filter from [1]
    input_coef = np.array(
        [
            -0.009341062898525,
            -0.025470289659360,
            -0.004235264826105,
            0.044152415456420,
            0.036493718347760,
            -0.011893961934740,
            -0.022917390623150,
            -0.006788163862310,
            0.000000000000000,
        ],
        dtype=np.float64,
    )
    output_coef = np.array(
        [
            1.00000000000000000000,
            -3.63367395910957000000,
            5.03689812757486000000,
            -3.09612247819666000000,
            0.50620507633883000000,
            0.32421701566682000000,
            -0.15685485875559000000,
            0.01949130205890000000,
            0.00000000000000000000,
        ],
        dtype=np.float64,
    )

    # scaling factor from [1]
    scaling_factor = (3 / 4096) / (2.6 / 256) * 237.5

    acceleration_30hz = computations.resample(acceleration, 1 / 30)

    initial_conditions = signal.lfilter_zi(input_coef, output_coef).reshape((-1, 1))

    acceleration_bpf, _ = signal.lfilter(
        input_coef,
        output_coef,
        acceleration_30hz.measurements,
        zi=np.repeat(
            initial_conditions, acceleration_30hz.measurements.shape[1], axis=-1
        )
        * acceleration_30hz.measurements[0],
        axis=0,
    )

    scaled_acceleration = acceleration_bpf * scaling_factor
    threshold_acceleration = np.floor(np.minimum(np.abs(scaled_acceleration), 128))
    threshold_acceleration[threshold_acceleration < 4] = 0

    acceleration_10hz = computations.resample(
        models.Measurement(
            measurements=threshold_acceleration, time=acceleration_30hz.time
        ),
        1 / 10,
    )
    acceleration_10hz.measurements = np.floor(acceleration_10hz.measurements)

    aggregator = pl.exclude(["time"]).drop_nans()
    epoch_counts = (
        acceleration_10hz.lazy_frame()
        .group_by_dynamic("time", every=f"{epoch_length_nanoseconds}ns")
        .agg(aggregator.sum())
    ).collect()

    ag_counts = epoch_counts.with_columns(
        (pl.col("column_0") ** 2 + pl.col("column_1") ** 2 + pl.col("column_2") ** 2)
        .sqrt()
        .alias("magnitude")
    )
    ag_counts = ag_counts.drop(["column_0", "column_1", "column_2"])

    return models.Measurement.from_data_frame(ag_counts)


def detect_nonwear(
    acceleration: models.Measurement,
    short_epoch_length: int = 900,
    n_short_epoch_in_long_epoch: int = 4,
    std_criteria: float = 0.013,
) -> models.Measurement:
    """Set non_wear_flag based on accelerometer data.

    This implements a modified version of the GGIR "2023" non-wear detection algorithm.
    Briefly, the algorithm, creates a sliding window of long epoch length that steps
    forward by the short epoch length. The long epoch length is an integer multiple of
    the short epoch length, that can be specified by the user.
    It checks if the acceleration data in that long window, for each axis, meets the
    criteria threshold for the standard deviation of acceleration values to
    compute a non-wear value. The total non-wear value (0, 1, 2, 3) for the long window
    is the sum of each axis.
    The non-wear value is applied to all the short windows that make up the long
    window. Additionally, as the majority of the short windows are part of multiple long
    windows, the value of a short window is updated to the maximum nonwear value from
    these overlaps.
    Finally, there is a pass to find isolated "1s" in the non-wear
    value, and set them to 2 if surrounded by > 1 values. The non-wear flag is set to
    1 (true) if the non-wear value is >= 2, and 0 (false) otherwise.


    Args:
        acceleration: The Measurement instance that contains the calibrated acceleration
            data.
        short_epoch_length: The short window size, in seconds.
        n_short_epoch_in_long_epoch: Number of short epochs that makeup one long epoch.
        std_criteria: Threshold criteria for standard deviation.


    Returns:
        A new Measurement instance with the non-wear flag and corresponding timestamps.
    """
    logger.debug("Detecting non-wear data.")
    acceleration_grouped_by_short_window = _group_acceleration_data_by_time(
        acceleration, short_epoch_length
    )

    nonwear_value_array = _compute_nonwear_value_array(
        acceleration_grouped_by_short_window,
        n_short_epoch_in_long_epoch,
        std_criteria,
    )

    nonwear_value_array_cleaned = _cleanup_isolated_ones_nonwear_value(
        nonwear_value_array
    )
    non_wear_flag = np.where(nonwear_value_array_cleaned >= 2, 1, 0)

    return models.Measurement(
        measurements=non_wear_flag, time=acceleration_grouped_by_short_window["time"]
    )


def combined_temp_accel_detect_nonwear(
    acceleration: models.Measurement,
    temperature: models.Measurement,
    std_criteria: float = 0.013,
    temperature_threshold: float = 26.0,
) -> models.Measurement:
    """This function uses the both temperature and acceleration data to detect non-wear.

    The function implements the algorithm described in the Zhou 2015 paper. Briefly,
    data is chunked into one minute windows. Each windows is classified as "wear" or
    "non-wear", here denoted by 0 and 1, respectively.

    Args:
        acceleration: The Measurement instance that contains calibrated acceleration.
        temperature: The Measurement instance that contains the temperature data.
        std_criteria: Threshold criteria for standard deviation.
        temperature_threshold: The temperature threshold for non-wear detection.

    Returns:
        A new Measurement instance with the non-wear flag and corresponding timestamps.
        The temporal resolutions is one minute.

    References:
        Zhou S, Hill RA, Morgan K, et al, Classification of accelerometer wear and
        non-wear events in seconds for monitoring free-living physical activityBMJ
        Open 2015; 5:e007447. doi: 10.1136/bmjopen-2014-007447.
    """
    logger.debug("Combined temperature and accelereation non-wear detection algorithm.")

    acceleration_grouped_by_window = _group_acceleration_data_by_time(acceleration, 60)
    low_pass_temp = _pre_process_temperature(temperature, acceleration)

    temperature_grouped_by_window = low_pass_temp.group_by_dynamic(
        index_column="time", every="60s"
    ).agg([pl.exclude(["time"])])

    total_n_short_windows = len(acceleration_grouped_by_window)

    nonwear_value_array = np.zeros(total_n_short_windows)

    for window_n in range(total_n_short_windows):
        mean_temp = temperature_grouped_by_window["temperature"][window_n].mean()
        axis_nonwear_value = (
            acceleration_grouped_by_window[window_n]
            .select(
                pl.col("X", "Y", "Z").map_batches(
                    lambda df: _compute_nonwear_value_per_axis(
                        df,
                        std_criteria,
                    )
                )
            )
            .sum_horizontal()
            .to_numpy()
        )
        if mean_temp < temperature_threshold and (axis_nonwear_value >= 2):
            nonwear_value_array[window_n] = 1
        elif window_n > 0 and mean_temp < temperature_threshold:
            mean_temp_previous = temperature_grouped_by_window["temperature"][
                window_n - 1
            ].mean()
            if mean_temp <= mean_temp_previous:
                nonwear_value_array[window_n] = 1

    return models.Measurement(
        measurements=nonwear_value_array,
        time=acceleration_grouped_by_window["time"],
    )


def detach_nonwear(
    acceleration: models.Measurement,
    temperature: models.Measurement,
    std_criteria: float = 0.013,
) -> models.Measurement:
    """This function implements the scikit DETACH algorithm for non-wear detection.

    The nonwear array is downsampled to 60 second resolution.

    Args:
        acceleration: The Measurement instance that contains calibrated acceleration.
        temperature: The Measurement instance that contains the temperature data.
        std_criteria: The acceleration STD threshold for non-wear detection.

    Returns:
        A new Measurement instance with the non-wear flag and corresponding timestamps.
        The temporal resolutions is one minute.

    References:
        A. Vert et al., “Detecting accelerometer non-wear periods using change
        in acceleration combined with rate-of-change in temperature,” BMC Medical
        Research Methodology, vol. 22, no. 1, p. 147, May 2022,
        doi: 10.1186/s12874-022-01633-6.
    """

    def upsample_temperature(
        temperature: np.ndarray, acceleration: np.ndarray
    ) -> np.ndarray:
        """Helper function to upsample the temperature to match acceleration data."""
        x_temperature = np.linspace(0, 1, len(temperature))
        x_acceleration = np.linspace(0, 1, len(acceleration))
        temperature_interpolator = interpolate.interp1d(x_temperature, temperature)
        return temperature_interpolator(x_acceleration)

    def cleanup_DETACH_nonwear(nonwear: models.Measurement) -> models.Measurement:
        """Helper function to downsample DETACH output to 60s."""
        nonwear_downsample = computations.moving_mean(nonwear, 60)

        nonwear_downsample.measurements = np.where(
            nonwear_downsample.measurements >= 0.5, 1, 0
        )

        return nonwear_downsample

    logger.debug("Running scikit DETACH algorithm.")
    time = acceleration.time.dt.timestamp(time_unit="ns").to_numpy()
    upsample_temp = upsample_temperature(
        temperature.measurements, acceleration.measurements
    )

    detach_class = wear_detection.DETACH(sd_thresh=std_criteria)
    nonwear_array = np.ones(len(time), dtype=np.float32)
    detach_wear = detach_class.predict(
        time=time / 1e9, accel=acceleration.measurements, temperature=upsample_temp
    )

    for start, stop in detach_wear["wear"]:
        nonwear_array[start:stop] = 0

    non_wear_time = readers.unix_epoch_time_to_polars_datetime(time)

    nonwear = models.Measurement(measurements=nonwear_array, time=non_wear_time)

    return cleanup_DETACH_nonwear(nonwear)


def _group_acceleration_data_by_time(
    acceleration: models.Measurement, window_length: int
) -> pl.DataFrame:
    """Helper function to group the acceleration data by short windows.

    Args:
        acceleration: The Measurement instance that contains calibrated acceleration.
        window_length: The window size, in seconds.

    Returns:
        A polars DataFrame with the acceleration data grouped by window_length.
    """
    acceleration_polars_df = pl.DataFrame(
        {
            "X": acceleration.measurements[:, 0],
            "Y": acceleration.measurements[:, 1],
            "Z": acceleration.measurements[:, 2],
            "time": acceleration.time,
        }
    )
    acceleration_polars_df = acceleration_polars_df.with_columns(
        pl.col("time").set_sorted()
    )

    acceleration_grouped_by_window_length = acceleration_polars_df.group_by_dynamic(
        index_column="time", every=(str(window_length) + "s")
    ).agg([pl.all().exclude(["time"])])

    return acceleration_grouped_by_window_length


def _compute_nonwear_value_array(
    grouped_acceleration: pl.DataFrame,
    n_short_epoch_in_long_epoch: int,
    std_criteria: float,
) -> np.ndarray:
    """Helper function to calculate the nonwear value array.

    This function calculates the nonwear value array based on the GGIR 2023 methodology.
    It computes the nonwear value for each axis, based on the acceleration data that
    makes up one long epoch window. That nonwear value is then applied to all the
    short windows that make up the long window. It iterates forward by one short_window
    length and repeats the process. For the overlapping short windows, the maximum
    nonwear value is kept and is assigned to the nonwear value array.

    Args:
        grouped_acceleration: The acceleration data grouped into short windows.
        n_short_epoch_in_long_epoch: Number of short epochs that makeup one long epoch.
        std_criteria: Threshold criteria for standard deviation.

    Returns:
        Non-wear value array.
    """
    total_n_short_windows = len(grouped_acceleration)
    nonwear_value_array = np.zeros(total_n_short_windows)

    for window_n in range(total_n_short_windows - n_short_epoch_in_long_epoch + 1):
        acceleration_selected_long_window = grouped_acceleration[
            window_n : window_n + n_short_epoch_in_long_epoch
        ]

        calculated_nonwear_value = acceleration_selected_long_window.select(
            pl.col("X", "Y", "Z").map_batches(
                lambda df: _compute_nonwear_value_per_axis(
                    df,
                    std_criteria,
                )
            )
        ).sum_horizontal()

        max_window_value = np.maximum(
            nonwear_value_array[window_n : window_n + n_short_epoch_in_long_epoch],
            np.repeat(calculated_nonwear_value, n_short_epoch_in_long_epoch),
        )
        nonwear_value_array[window_n : window_n + n_short_epoch_in_long_epoch] = (
            max_window_value
        )

    return nonwear_value_array


def _compute_nonwear_value_per_axis(
    axis_acceleration_data: pl.Series,
    std_criteria: float,
) -> bool:
    """Helper function to calculate the nonwear criteria per axis.

    Args:
        axis_acceleration_data: The long window acceleration data for one axis.
            It is a pl.Series chunked into short windows where each row is a list of the
            acceleration data of one axis (length of each list is the number of samples
            that make up short_epoch_length in seconds).
        std_criteria: Threshold criteria for standard deviation


    Returns:
        Non-wear value for the axis.
    """
    axis_long_window_data = pl.concat(axis_acceleration_data, how="vertical")
    axis_std = axis_long_window_data.std()
    criteria_boolean = axis_std < std_criteria

    return criteria_boolean


def _cleanup_isolated_ones_nonwear_value(nonwear_value_array: np.ndarray) -> np.ndarray:
    """Helper function to clean up isolated ones in nonwear value array.

    This function finds isolated ones in the nonwear value array and
    sets them to 2 if they are surrounded by values > 1.

    Args:
        nonwear_value_array: The nonwear value array that needs to be cleaned up.
            It is a 1D numpy array.

    Returns:
        The modified nonwear value array.
    """
    nonwear_value_array = nonwear_value_array.astype(int)

    left_neighbors = np.roll(nonwear_value_array, 1)
    right_neighbors = np.roll(nonwear_value_array, -1)

    condition = (left_neighbors > 1) & (right_neighbors > 1) & nonwear_value_array == 1
    condition[0] = False
    condition[-1] = False

    nonwear_value_array[condition] = 2

    return nonwear_value_array


def _pre_process_temperature(
    temperature: models.Measurement, acceleration: models.Measurement
) -> pl.DataFrame:
    """Pre-process temperature data for non-wear detection.

    This function is used specificall for the CTA, it ensures that the temperature
    and acceleration Measurements end at the same time (due to the inherent sampling
    rate differences on the device), upsamples the temperature data to 1 second
    intervals and then low-pass filters the data by applying a 2 minute rolling mean.

    Args:
        temperature: The Measurement instance that contains the temperature data.
        acceleration: The Measurement instance that contains the acceleration data.

    Returns:
        A polars DataFrame with the pre-processed temperature data.

    Raises:
        ValueError: If the acceleration data ends before the temperature data.
    """
    temperature_polars_df = pl.DataFrame(
        {
            "time": temperature.time,
            "temperature": temperature.measurements,
        }
    )

    if acceleration.time[-1] > temperature.time[-1]:
        new_row = pl.DataFrame(
            {
                "time": acceleration.time[-1],
                "temperature": temperature.measurements[-1],
            },
            schema={"time": pl.Datetime("ns"), "temperature": pl.Float32},
        )

        temperature_polars_df = temperature_polars_df.vstack(new_row)
    elif temperature.time[-1] > acceleration.time[-1]:
        raise ValueError("Temperature data ends before acceleration data.")

    temperature_polars_df = temperature_polars_df.with_columns(
        pl.col("time").set_sorted()
    )
    upsample_temp = temperature_polars_df.upsample(
        time_column="time", every="1s", maintain_order=True
    ).interpolate()
    upsample_temp = upsample_temp.with_columns(
        pl.col("temperature").fill_null(strategy="forward")
    )

    low_pass_temp = upsample_temp.rolling(index_column="time", period="120s").agg(
        [pl.mean("temperature")]
    )
    return low_pass_temp


def monitor_independent_movement_summary_units(
    acceleration: models.Measurement,
    combination_method: Literal["sum", "vector_magnitude"] = "sum",
    epoch: float = 60.0,
    interpolation_frequency: int = 100,
    dynamic_range: tuple[float, float] = (-8.0, 8.0),
    cutoffs: Tuple[float, float] = (0.2, 5.0),
    order: int = 4,
    *,
    rectify: bool = True,
) -> models.Measurement:
    """Calculates monitor independent movement summary units (MIMS).

    This function processes raw acceleration data to calculate MIMS units,
    based on the original R implementation by John et al.(2019). The main processing
    steps as described in  the original paper are interpolation (100Hz by default),
    extrapolation of any values that went beyond the dynamic range of the device,
    filtering using a 4th order butterworth filter, aggregation by calculating the area
    under the curve over a given epoch, and finally truncation of small values.
    The MIMS value per axis is then combined through a sum or vector magnitude, and
    returned as a single vector.

    Args:
        acceleration: Triaxial acceleration data to be processed.
        combination_method: Method to combine MIMS values across axes.
        epoch: Duration over which each MIMS value will be calculated. Measured in
            seconds.
        interpolation_frequency: Frequency to interpolate acceleration data, defaults
            to 100 Hz as described in John et al. 2019.
        dynamic_range: Tuple specifying the minimum and maximum values (in g) of the
            accelerometer's dynamic range.
        cutoffs: Tuple specifying the low and high cutoff frequencies (in Hz) for the
            Butterworth filter.
        order: Order of the Butterworth filter.
        rectify: Specifies if data should be rectified before integration. If True any
            value below -150 will assign the value of that axis to -1 for that epoch.
            Additionally the absolute value of accelerometer data will be used for
            integration.

    Returns:
        Processed MIMS values after combining the values of each axis with the provided
        method.

    References:
        John, D., Tang, Q., Albinali, F. and Intille, S., 2019. An Open-Source
        Monitor-Independent Movement Summary for Accelerometer Data Processing. Journal
        for the Measurement of Physical Behaviour, 2(4), pp.268-281.

    """
    interpolated_measure = mims.interpolate_measure(
        acceleration=acceleration,
        new_frequency=interpolation_frequency,
    )
    extrapolated_measure = mims.extrapolate_points(
        acceleration=interpolated_measure,
        dynamic_range=dynamic_range,
        sampling_rate=interpolation_frequency,
    )
    filtered_measure = mims.butterworth_filter(
        acceleration=extrapolated_measure,
        sampling_rate=interpolation_frequency,
        cutoffs=cutoffs,
        order=order,
    )
    aggregated_measure = mims.aggregate_mims(
        acceleration=filtered_measure,
        epoch=epoch,
        sampling_rate=interpolation_frequency,
        rectify=rectify,
    )
    combined_mims = mims.combine_mims(
        acceleration=aggregated_measure, combination_method=combination_method
    )
    return combined_mims



"""Internal data model."""

from typing import Optional

import numpy as np
import polars as pl
from pydantic import BaseModel, field_validator

from wristpy.core import config

VALID_FILE_TYPES = (".csv", ".parquet")

logger = config.get_logger()


class Measurement(BaseModel):
    """A single measurement of a sensor and its corresponding time."""

    measurements: np.ndarray
    time: pl.Series

    @classmethod
    def from_data_frame(cls, data_frame: pl.DataFrame) -> "Measurement":
        """Creates a measurement from a Polars DataFrame.

        Args:
            data_frame: The Polars DataFrame, must have a time column. All
                non-time columns will be used as the 'measurements' input.
        """
        return Measurement(
            measurements=data_frame.drop("time").to_numpy().squeeze(),
            time=data_frame["time"],
        )

    def lazy_frame(self) -> pl.LazyFrame:
        """Converts the measurement to a LazyFrame.

        Returns:
            The Measurement as a LazyFrame. The time property will have column name
                'time'. Other column names should not be relied upon.
        """
        return pl.concat(
            [
                pl.LazyFrame(self.measurements),
                pl.LazyFrame({"time": self.time}).set_sorted("time"),
            ],
            how="horizontal",
        )

    class Config:
        """Config to allow for ndarray as input."""

        arbitrary_types_allowed = True

    @field_validator("measurements")
    def validate_measurements_not_empty(cls, v: np.ndarray) -> np.ndarray:
        """Validate that the measurements array is not empty.

        Args:
            cls: The class.
            v: The measurements array to validate.

        Returns:
            v: The measurements array if it is not empty.

        Raises:
            ValueError: If the measurements array is empty.
        """
        if v.size == 0:
            raise ValueError("measurements array must not be empty")
        return v

    @field_validator("time")
    def validate_time(cls, v: pl.Series) -> pl.Series:
        """Validate the time series.

        Check that the time series is a datetime series, contains only unique entries,
        and is sorted.

        Args:
            cls: The class.
            v: The time series to validate.

        Returns:
            v: The time series if it is valid.

        Raises:
            ValueError: If the time series is not a datetime series or is not sorted,
            or is empty.
        """
        if not isinstance(v.dtype, pl.datatypes.Datetime):
            raise ValueError("Time must be a datetime series")
        if not v.is_unique().all():
            raise ValueError("Time series must contain unique entries")
        if not v.is_sorted():
            raise ValueError("Time series must be sorted")
        if v.is_empty():
            raise ValueError("Time series cannot be empty")
        return v


class WatchData(BaseModel):
    """Watch data that is read off the device.

    This class should provide access to all raw input data.
    It must not be mutated during processing.
    """

    acceleration: Measurement
    lux: Optional[Measurement] = None
    battery: Optional[Measurement] = None
    capsense: Optional[Measurement] = None
    temperature: Optional[Measurement] = None
    idle_sleep_mode_flag: Optional[bool] = None
    dynamic_range: Optional[tuple[float, float]] = None

    @field_validator("acceleration")
    def validate_acceleration(cls, v: Measurement) -> Measurement:
        """Validate the acceleration data.

        Ensure that the acceleration data is a 2D array with 3 columns.

        Args:
            cls: The class.
            v: The acceleration data to validate.

        Returns:
            v: The acceleration data if it is valid.

        Raises:
            ValueError: If the acceleration data is not a 2D array with 3 columns.
        """
        if v.measurements.ndim != 2 or v.measurements.shape[1] != 3:
            raise ValueError("acceleration must be a 2D array with 3 columns")
        return v
